import json
import logging
import os
import platform
from abc import ABCMeta, abstractmethod
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Container,
    Dict,
    List,
    Mapping,
    Optional,
    Tuple,
    Type,
    Union,
)

import gymnasium as gym
import numpy as np
import tree  # pip install dm_tree
from gymnasium.spaces import Box
from packaging import version

import ray
import ray.cloudpickle as pickle
from ray.actor import ActorHandle
from ray.air.checkpoint import Checkpoint
from ray.rllib.core.models.base import STATE_IN, STATE_OUT
from ray.rllib.models.action_dist import ActionDistribution
from ray.rllib.models.catalog import ModelCatalog
from ray.rllib.models.modelv2 import ModelV2
from ray.rllib.policy.rnn_sequencing import add_time_dimension
from ray.rllib.policy.sample_batch import SampleBatch
from ray.rllib.policy.view_requirement import ViewRequirement
from ray.rllib.utils.annotations import (
    DeveloperAPI,
    ExperimentalAPI,
    OverrideToImplementCustomLogic,
    OverrideToImplementCustomLogic_CallToSuperRecommended,
    is_overridden,
)
from ray.rllib.utils.checkpoints import (
    CHECKPOINT_VERSION,
    get_checkpoint_info,
    try_import_msgpack,
)
from ray.rllib.utils.deprecation import (
    Deprecated,
    DEPRECATED_VALUE,
    deprecation_warning,
)
from ray.rllib.utils.exploration.exploration import Exploration
from ray.rllib.utils.framework import try_import_tf, try_import_torch
from ray.rllib.utils.from_config import from_config
from ray.rllib.utils.numpy import convert_to_numpy
from ray.rllib.utils.serialization import (
    deserialize_type,
    space_from_dict,
    space_to_dict,
)
from ray.rllib.utils.spaces.space_utils import (
    get_base_struct_from_space,
    get_dummy_batch_for_space,
    unbatch,
)
from ray.rllib.utils.tensor_dtype import get_np_dtype
from ray.rllib.utils.tf_utils import get_tf_eager_cls_if_necessary
from ray.rllib.utils.typing import (
    AgentID,
    AlgorithmConfigDict,
    ModelGradients,
    ModelWeights,
    PolicyID,
    PolicyState,
    T,
    TensorStructType,
    TensorType,
)
from ray.util.annotations import PublicAPI

tf1, tf, tfv = try_import_tf()
torch, _ = try_import_torch()

if TYPE_CHECKING:
    from ray.rllib.evaluation import Episode
    from ray.rllib.core.rl_module import RLModule


logger = logging.getLogger(__name__)


@PublicAPI
class PolicySpec:
    """A policy spec used in the "config.multiagent.policies" specification dict.

    As values (keys are the policy IDs (str)). E.g.:
    config:
        multiagent:
            policies: {
                "pol1": PolicySpec(None, Box, Discrete(2), {"lr": 0.0001}),
                "pol2": PolicySpec(config={"lr": 0.001}),
            }
    """

    def __init__(
        self, policy_class=None, observation_space=None, action_space=None, config=None
    ):
        # If None, use the Algorithm's default policy class stored under
        # `Algorithm._policy_class`.
        self.policy_class = policy_class
        # If None, use the env's observation space. If None and there is no Env
        # (e.g. offline RL), an error is thrown.
        self.observation_space = observation_space
        # If None, use the env's action space. If None and there is no Env
        # (e.g. offline RL), an error is thrown.
        self.action_space = action_space
        # Overrides defined keys in the main Algorithm config.
        # If None, use {}.
        self.config = config

    def __eq__(self, other: "PolicySpec"):
        return (
            self.policy_class == other.policy_class
            and self.observation_space == other.observation_space
            and self.action_space == other.action_space
            and self.config == other.config
        )

    def serialize(self) -> Dict:
        from ray.rllib.algorithms.registry import get_policy_class_name

        # Try to figure out a durable name for this policy.
        cls = get_policy_class_name(self.policy_class)
        if cls is None:
            logger.warning(
                f"Can not figure out a durable policy name for {self.policy_class}. "
                f"You are probably trying to checkpoint a custom policy. "
                f"Raw policy class may cause problems when the checkpoint needs to "
                "be loaded in the future. To fix this, make sure you add your "
                "custom policy in rllib.algorithms.registry.POLICIES."
            )
            cls = self.policy_class

        return {
            "policy_class": cls,
            "observation_space": space_to_dict(self.observation_space),
            "action_space": space_to_dict(self.action_space),
            # TODO(jungong) : try making the config dict durable by maybe
            #  getting rid of all the fields that are not JSON serializable.
            "config": self.config,
        }

    @classmethod
    def deserialize(cls, spec: Dict) -> "PolicySpec":
        if isinstance(spec["policy_class"], str):
            # Try to recover the actual policy class from durable name.
            from ray.rllib.algorithms.registry import get_policy_class

            policy_class = get_policy_class(spec["policy_class"])
        elif isinstance(spec["policy_class"], type):
            # Policy spec is already a class type. Simply use it.
            policy_class = spec["policy_class"]
        else:
            raise AttributeError(f"Unknown policy class spec {spec['policy_class']}")

        return cls(
            policy_class=policy_class,
            observation_space=space_from_dict(spec["observation_space"]),
            action_space=space_from_dict(spec["action_space"]),
            config=spec["config"],
        )


@DeveloperAPI
class Policy(metaclass=ABCMeta):
    """RLlib's base class for all Policy implementations.

    Policy is the abstract superclass for all DL-framework specific sub-classes
    (e.g. TFPolicy or TorchPolicy). It exposes APIs to

    1. Compute actions from observation (and possibly other) inputs.

    2. Manage the Policy's NN model(s), like exporting and loading their weights.

    3. Postprocess a given trajectory from the environment or other input via the
        `postprocess_trajectory` method.

    4. Compute losses from a train batch.

    5. Perform updates from a train batch on the NN-models (this normally includes loss
        calculations) either:

        a. in one monolithic step (`learn_on_batch`)

        b. via batch pre-loading, then n steps of actual loss computations  and updates
            (`load_batch_into_buffer` + `learn_on_loaded_batch`).
    """

    @DeveloperAPI
    def __init__(
        self,
        observation_space: gym.Space,
        action_space: gym.Space,
        config: AlgorithmConfigDict,
    ):
        """Initializes a Policy instance.

        Args:
            observation_space: Observation space of the policy.
            action_space: Action space of the policy.
            config: A complete Algorithm/Policy config dict. For the default
                config keys and values, see rllib/algorithm/algorithm.py.
        """
        self.observation_space: gym.Space = observation_space
        self.action_space: gym.Space = action_space
        # the policy id in the global context.
        self.__policy_id = config.get("__policy_id")
        # The base struct of the observation/action spaces.
        # E.g. action-space = gym.spaces.Dict({"a": Discrete(2)}) ->
        # action_space_struct = {"a": Discrete(2)}
        self.observation_space_struct = get_base_struct_from_space(observation_space)
        self.action_space_struct = get_base_struct_from_space(action_space)

        self.config: AlgorithmConfigDict = config
        self.framework = self.config.get("framework")

        # Create the callbacks object to use for handling custom callbacks.
        from ray.rllib.algorithms.callbacks import DefaultCallbacks

        callbacks = self.config.get("callbacks")
        if isinstance(callbacks, DefaultCallbacks):
            self.callbacks = callbacks()
        elif isinstance(callbacks, (str, type)):
            try:
                self.callbacks: "DefaultCallbacks" = deserialize_type(
                    self.config.get("callbacks")
                )()
            except Exception:
                pass  # TEST
        else:
            self.callbacks: "DefaultCallbacks" = DefaultCallbacks()

        # The global timestep, broadcast down from time to time from the
        # local worker to all remote workers.
        self.global_timestep: int = 0
        # The number of gradient updates this policy has undergone.
        self.num_grad_updates: int = 0

        # The action distribution class to use for action sampling, if any.
        # Child classes may set this.
        self.dist_class: Optional[Type] = None

        # Initialize view requirements.
        self.init_view_requirements()

        # Whether the Model's initial state (method) has been added
        # automatically based on the given view requirements of the model.
        self._model_init_state_automatically_added = False

        # Connectors.
        self.agent_connectors = None
        self.action_connectors = None

    @staticmethod
    def from_checkpoint(
        checkpoint: Union[str, Checkpoint],
        policy_ids: Optional[Container[PolicyID]] = None,
    ) -> Union["Policy", Dict[PolicyID, "Policy"]]:
        """Creates new Policy instance(s) from a given Policy or Algorithm checkpoint.

        Note: This method must remain backward compatible from 2.1.0 on, wrt.
        checkpoints created with Ray 2.0.0 or later.

        Args:
            checkpoint: The path (str) to a Policy or Algorithm checkpoint directory
                or an AIR Checkpoint (Policy or Algorithm) instance to restore
                from.
                If checkpoint is a Policy checkpoint, `policy_ids` must be None
                and only the Policy in that checkpoint is restored and returned.
                If checkpoint is an Algorithm checkpoint and `policy_ids` is None,
                will return a list of all Policy objects found in
                the checkpoint, otherwise a list of those policies in `policy_ids`.
            policy_ids: List of policy IDs to extract from a given Algorithm checkpoint.
                If None and an Algorithm checkpoint is provided, will restore all
                policies found in that checkpoint. If a Policy checkpoint is given,
                this arg must be None.

        Returns:
            An instantiated Policy, if `checkpoint` is a Policy checkpoint. A dict
            mapping PolicyID to Policies, if `checkpoint` is an Algorithm checkpoint.
            In the latter case, returns all policies within the Algorithm if
            `policy_ids` is None, else a dict of only those Policies that are in
            `policy_ids`.
        """
        checkpoint_info = get_checkpoint_info(checkpoint)

        # Algorithm checkpoint: Extract one or more policies from it and return them
        # in a dict (mapping PolicyID to Policy instances).
        if checkpoint_info["type"] == "Algorithm":
            from ray.rllib.algorithms.algorithm import Algorithm

            policies = {}

            # Old Algorithm checkpoints: State must be completely retrieved from:
            # algo state file -> worker -> "state".
            if checkpoint_info["checkpoint_version"] < version.Version("1.0"):
                with open(checkpoint_info["state_file"], "rb") as f:
                    state = pickle.load(f)
                # In older checkpoint versions, the policy states are stored under
                # "state" within the worker state (which is pickled in itself).
                worker_state = pickle.loads(state["worker"])
                policy_states = worker_state["state"]
                for pid, policy_state in policy_states.items():
                    # Get spec and config, merge config with
                    serialized_policy_spec = worker_state["policy_specs"][pid]
                    policy_config = Algorithm.merge_algorithm_configs(
                        worker_state["policy_config"], serialized_policy_spec["config"]
                    )
                    serialized_policy_spec.update({"config": policy_config})
                    policy_state.update({"policy_spec": serialized_policy_spec})
                    policies[pid] = Policy.from_state(policy_state)
            # Newer versions: Get policy states from "policies/" sub-dirs.
            elif checkpoint_info["policy_ids"] is not None:
                for policy_id in checkpoint_info["policy_ids"]:
                    if policy_ids is None or policy_id in policy_ids:
                        policy_checkpoint_info = get_checkpoint_info(
                            os.path.join(
                                checkpoint_info["checkpoint_dir"],
                                "policies",
                                policy_id,
                            )
                        )
                        assert policy_checkpoint_info["type"] == "Policy"
                        with open(policy_checkpoint_info["state_file"], "rb") as f:
                            policy_state = pickle.load(f)
                        policies[policy_id] = Policy.from_state(policy_state)
            return policies

        # Policy checkpoint: Return a single Policy instance.
        else:
            with open(checkpoint_info["state_file"], "rb") as f:
                state = pickle.load(f)
            return Policy.from_state(state)

    @staticmethod
    def from_state(state: PolicyState) -> "Policy":
        """Recovers a Policy from a state object.

        The `state` of an instantiated Policy can be retrieved by calling its
        `get_state` method. This only works for the V2 Policy classes (EagerTFPolicyV2,
        SynamicTFPolicyV2, and TorchPolicyV2). It contains all information necessary
        to create the Policy. No access to the original code (e.g. configs, knowledge of
        the policy's class, etc..) is needed.

        Args:
            state: The state to recover a new Policy instance from.

        Returns:
            A new Policy instance.
        """
        serialized_pol_spec: Optional[dict] = state.get("policy_spec")
        if serialized_pol_spec is None:
            raise ValueError(
                "No `policy_spec` key was found in given `state`! "
                "Cannot create new Policy."
            )
        pol_spec = PolicySpec.deserialize(serialized_pol_spec)
        actual_class = get_tf_eager_cls_if_necessary(
            pol_spec.policy_class,
            pol_spec.config,
        )

        if pol_spec.config["framework"] == "tf":
            from ray.rllib.policy.tf_policy import TFPolicy

            return TFPolicy._tf1_from_state_helper(state)

        # Create the new policy.
        new_policy = actual_class(
            # Note(jungong) : we are intentionally not using keyward arguments here
            # because some policies name the observation space parameter obs_space,
            # and some others name it observation_space.
            pol_spec.observation_space,
            pol_spec.action_space,
            pol_spec.config,
        )

        # Set the new policy's state (weights, optimizer vars, exploration state,
        # etc..).
        new_policy.set_state(state)
        # Return the new policy.
        return new_policy

    @ExperimentalAPI
    @OverrideToImplementCustomLogic
    def make_rl_module(self) -> "RLModule":
        """Returns the RL Module (only for when RLModule API is enabled.)

        If RLModule API is enabled (self.config.rl_module(_enable_rl_module_api=True),
        this method should be implemented and should return the RLModule instance to
        use for this Policy. Otherwise, RLlib will error out.
        """
        # if imported on top it creates circular dependency
        from ray.rllib.core.rl_module.rl_module import SingleAgentRLModuleSpec

        if self.__policy_id is None:
            raise ValueError(
                "When using RLModule API, `policy_id` within the policies must be "
                "set. This should have happened automatically. If you see this "
                "bug, please file a github issue."
            )

        spec = self.config["__marl_module_spec"]
        if isinstance(spec, SingleAgentRLModuleSpec):
            module = spec.build()
        else:
            # filter the module_spec to only contain the policy_id of this policy
            marl_spec = type(spec)(
                marl_module_class=spec.marl_module_class,
                module_specs={self.__policy_id: spec.module_specs[self.__policy_id]},
            )
            marl_module = marl_spec.build()
            module = marl_module[self.__policy_id]

        return module

    @DeveloperAPI
    def init_view_requirements(self):
        """Maximal view requirements dict for `learn_on_batch()` and
        `compute_actions` calls.
        Specific policies can override this function to provide custom
        list of view requirements.
        """
        # Maximal view requirements dict for `learn_on_batch()` and
        # `compute_actions` calls.
        # View requirements will be automatically filtered out later based
        # on the postprocessing and loss functions to ensure optimal data
        # collection and transfer performance.
        view_reqs = self._get_default_view_requirements()
        if not hasattr(self, "view_requirements"):
            self.view_requirements = view_reqs
        else:
            for k, v in view_reqs.items():
                if k not in self.view_requirements:
                    self.view_requirements[k] = v

    def get_connector_metrics(self) -> Dict:
        """Get metrics on timing from connectors."""
        return {
            "agent_connectors": {
                name + "_ms": 1000 * timer.mean
                for name, timer in self.agent_connectors.timers.items()
            },
            "action_connectors": {
                name + "_ms": 1000 * timer.mean
                for name, timer in self.agent_connectors.timers.items()
            },
        }

    def reset_connectors(self, env_id) -> None:
        """Reset action- and agent-connectors for this policy."""
        self.agent_connectors.reset(env_id=env_id)
        self.action_connectors.reset(env_id=env_id)

    @DeveloperAPI
    def compute_single_action(
        self,
        obs: Optional[TensorStructType] = None,
        state: Optional[List[TensorType]] = None,
        *,
        prev_action: Optional[TensorStructType] = None,
        prev_reward: Optional[TensorStructType] = None,
        info: dict = None,
        input_dict: Optional[SampleBatch] = None,
        episode: Optional["Episode"] = None,
        explore: Optional[bool] = None,
        timestep: Optional[int] = None,
        # Kwars placeholder for future compatibility.
        **kwargs,
    ) -> Tuple[TensorStructType, List[TensorType], Dict[str, TensorType]]:
        """Computes and returns a single (B=1) action value.

        Takes an input dict (usually a SampleBatch) as its main data input.
        This allows for using this method in case a more complex input pattern
        (view requirements) is needed, for example when the Model requires the
        last n observations, the last m actions/rewards, or a combination
        of any of these.
        Alternatively, in case no complex inputs are required, takes a single
        `obs` values (and possibly single state values, prev-action/reward
        values, etc..).

        Args:
            obs: Single observation.
            state: List of RNN state inputs, if any.
            prev_action: Previous action value, if any.
            prev_reward: Previous reward, if any.
            info: Info object, if any.
            input_dict: A SampleBatch or input dict containing the
                single (unbatched) Tensors to compute actions. If given, it'll
                be used instead of `obs`, `state`, `prev_action|reward`, and
                `info`.
            episode: This provides access to all of the internal episode state,
                which may be useful for model-based or multi-agent algorithms.
            explore: Whether to pick an exploitation or
                exploration action
                (default: None -> use self.config["explore"]).
            timestep: The current (sampling) time step.

        Keyword Args:
            kwargs: Forward compatibility placeholder.

        Returns:
            Tuple consisting of the action, the list of RNN state outputs (if
            any), and a dictionary of extra features (if any).
        """
        # Build the input-dict used for the call to
        # `self.compute_actions_from_input_dict()`.
        if input_dict is None:
            input_dict = {SampleBatch.OBS: obs}
            if state is not None:
                if self.config.get("_enable_rl_module_api", False):
                    input_dict["state_in"] = state
                else:
                    for i, s in enumerate(state):
                        input_dict[f"state_in_{i}"] = s
            if prev_action is not None:
                input_dict[SampleBatch.PREV_ACTIONS] = prev_action
            if prev_reward is not None:
                input_dict[SampleBatch.PREV_REWARDS] = prev_reward
            if info is not None:
                input_dict[SampleBatch.INFOS] = info

        # Batch all data in input dict.
        input_dict = tree.map_structure_with_path(
            lambda p, s: (
                s
                if p == "seq_lens"
                else s.unsqueeze(0)
                if torch and isinstance(s, torch.Tensor)
                else np.expand_dims(s, 0)
            ),
            input_dict,
        )

        episodes = None
        if episode is not None:
            episodes = [episode]

        out = self.compute_actions_from_input_dict(
            input_dict=SampleBatch(input_dict),
            episodes=episodes,
            explore=explore,
            timestep=timestep,
        )

        # Some policies don't return a tuple, but always just a single action.
        # E.g. ES and ARS.
        if not isinstance(out, tuple):
            single_action = out
            state_out = []
            info = {}
        # Normal case: Policy should return (action, state, info) tuple.
        else:
            batched_action, state_out, info = out
            single_action = unbatch(batched_action)
        assert len(single_action) == 1
        single_action = single_action[0]

        # Return action, internal state(s), infos.
        return (
            single_action,
            tree.map_structure(lambda x: x[0], state_out),
            tree.map_structure(lambda x: x[0], info),
        )

    @DeveloperAPI
    def compute_actions_from_input_dict(
        self,
        input_dict: Union[SampleBatch, Dict[str, TensorStructType]],
        explore: Optional[bool] = None,
        timestep: Optional[int] = None,
        episodes: Optional[List["Episode"]] = None,
        **kwargs,
    ) -> Tuple[TensorType, List[TensorType], Dict[str, TensorType]]:
        """Computes actions from collected samples (across multiple-agents).

        Takes an input dict (usually a SampleBatch) as its main data input.
        This allows for using this method in case a more complex input pattern
        (view requirements) is needed, for example when the Model requires the
        last n observations, the last m actions/rewards, or a combination
        of any of these.

        Args:
            input_dict: A SampleBatch or input dict containing the Tensors
                to compute actions. `input_dict` already abides to the
                Policy's as well as the Model's view requirements and can
                thus be passed to the Model as-is.
            explore: Whether to pick an exploitation or exploration
                action (default: None -> use self.config["explore"]).
            timestep: The current (sampling) time step.
            episodes: This provides access to all of the internal episodes'
                state, which may be useful for model-based or multi-agent
                algorithms.

        Keyword Args:
            kwargs: Forward compatibility placeholder.

        Returns:
            actions: Batch of output actions, with shape like
                [BATCH_SIZE, ACTION_SHAPE].
            state_outs: List of RNN state output
                batches, if any, each with shape [BATCH_SIZE, STATE_SIZE].
            info: Dictionary of extra feature batches, if any, with shape like
                {"f1": [BATCH_SIZE, ...], "f2": [BATCH_SIZE, ...]}.
        """
        # Default implementation just passes obs, prev-a/r, and states on to
        # `self.compute_actions()`.
        state_batches = [s for k, s in input_dict.items() if k.startswith("state_in")]
        return self.compute_actions(
            input_dict[SampleBatch.OBS],
            state_batches,
            prev_action_batch=input_dict.get(SampleBatch.PREV_ACTIONS),
            prev_reward_batch=input_dict.get(SampleBatch.PREV_REWARDS),
            info_batch=input_dict.get(SampleBatch.INFOS),
            explore=explore,
            timestep=timestep,
            episodes=episodes,
            **kwargs,
        )

    @abstractmethod
    @DeveloperAPI
    def compute_actions(
        self,
        obs_batch: Union[List[TensorStructType], TensorStructType],
        state_batches: Optional[List[TensorType]] = None,
        prev_action_batch: Union[List[TensorStructType], TensorStructType] = None,
        prev_reward_batch: Union[List[TensorStructType], TensorStructType] = None,
        info_batch: Optional[Dict[str, list]] = None,
        episodes: Optional[List["Episode"]] = None,
        explore: Optional[bool] = None,
        timestep: Optional[int] = None,
        **kwargs,
    ) -> Tuple[TensorType, List[TensorType], Dict[str, TensorType]]:
        """Computes actions for the current policy.

        Args:
            obs_batch: Batch of observations.
            state_batches: List of RNN state input batches, if any.
            prev_action_batch: Batch of previous action values.
            prev_reward_batch: Batch of previous rewards.
            info_batch: Batch of info objects.
            episodes: List of Episode objects, one for each obs in
                obs_batch. This provides access to all of the internal
                episode state, which may be useful for model-based or
                multi-agent algorithms.
            explore: Whether to pick an exploitation or exploration action.
                Set to None (default) for using the value of
                `self.config["explore"]`.
            timestep: The current (sampling) time step.

        Keyword Args:
            kwargs: Forward compatibility placeholder

        Returns:
            actions: Batch of output actions, with shape like
                [BATCH_SIZE, ACTION_SHAPE].
            state_outs (List[TensorType]): List of RNN state output
                batches, if any, each with shape [BATCH_SIZE, STATE_SIZE].
            info (List[dict]): Dictionary of extra feature batches, if any,
                with shape like
                {"f1": [BATCH_SIZE, ...], "f2": [BATCH_SIZE, ...]}.
        """
        raise NotImplementedError

    @DeveloperAPI
    def compute_log_likelihoods(
        self,
        actions: Union[List[TensorType], TensorType],
        obs_batch: Union[List[TensorType], TensorType],
        state_batches: Optional[List[TensorType]] = None,
        prev_action_batch: Optional[Union[List[TensorType], TensorType]] = None,
        prev_reward_batch: Optional[Union[List[TensorType], TensorType]] = None,
        actions_normalized: bool = True,
        in_training: bool = True,
    ) -> TensorType:
        """Computes the log-prob/likelihood for a given action and observation.

        The log-likelihood is calculated using this Policy's action
        distribution class (self.dist_class).

        Args:
            actions: Batch of actions, for which to retrieve the
                log-probs/likelihoods (given all other inputs: obs,
                states, ..).
            obs_batch: Batch of observations.
            state_batches: List of RNN state input batches, if any.
            prev_action_batch: Batch of previous action values.
            prev_reward_batch: Batch of previous rewards.
            actions_normalized: Is the given `actions` already normalized
                (between -1.0 and 1.0) or not? If not and
                `normalize_actions=True`, we need to normalize the given
                actions first, before calculating log likelihoods.
            in_training: Whether to use the forward_train() or forward_exploration() of
                the underlying RLModule.
        Returns:
            Batch of log probs/likelihoods, with shape: [BATCH_SIZE].
        """
        raise NotImplementedError

    @DeveloperAPI
    @OverrideToImplementCustomLogic_CallToSuperRecommended
    def postprocess_trajectory(
        self,
        sample_batch: SampleBatch,
        other_agent_batches: Optional[
            Dict[AgentID, Tuple["Policy", SampleBatch]]
        ] = None,
        episode: Optional["Episode"] = None,
    ) -> SampleBatch:
        """Implements algorithm-specific trajectory postprocessing.

        This will be called on each trajectory fragment computed during policy
        evaluation. Each fragment is guaranteed to be only from one episode.
        The given fragment may or may not contain the end of this episode,
        depending on the `batch_mode=truncate_episodes|complete_episodes`,
        `rollout_fragment_length`, and other settings.

        Args:
            sample_batch: batch of experiences for the policy,
                which will contain at most one episode trajectory.
            other_agent_batches: In a multi-agent env, this contains a
                mapping of agent ids to (policy, agent_batch) tuples
                containing the policy and experiences of the other agents.
            episode: An optional multi-agent episode object to provide
                access to all of the internal episode state, which may
                be useful for model-based or multi-agent algorithms.

        Returns:
            The postprocessed sample batch.
        """
        # The default implementation just returns the same, unaltered batch.
        return sample_batch

    @ExperimentalAPI
    @OverrideToImplementCustomLogic
    def loss(
        self, model: ModelV2, dist_class: ActionDistribution, train_batch: SampleBatch
    ) -> Union[TensorType, List[TensorType]]:
        """Loss function for this Policy.

        Override this method in order to implement custom loss computations.

        Args:
            model: The model to calculate the loss(es).
            dist_class: The action distribution class to sample actions
                from the model's outputs.
            train_batch: The input batch on which to calculate the loss.

        Returns:
            Either a single loss tensor or a list of loss tensors.
        """
        raise NotImplementedError

    @DeveloperAPI
    def learn_on_batch(self, samples: SampleBatch) -> Dict[str, TensorType]:
        """Perform one learning update, given `samples`.

        Either this method or the combination of `compute_gradients` and
        `apply_gradients` must be implemented by subclasses.

        Args:
            samples: The SampleBatch object to learn from.

        Returns:
            Dictionary of extra metadata from `compute_gradients()`.

        Examples:
            >>> policy, sample_batch = ... # doctest: +SKIP
            >>> policy.learn_on_batch(sample_batch) # doctest: +SKIP
        """
        # The default implementation is simply a fused `compute_gradients` plus
        # `apply_gradients` call.
        grads, grad_info = self.compute_gradients(samples)
        self.apply_gradients(grads)
        return grad_info

    @ExperimentalAPI
    def learn_on_batch_from_replay_buffer(
        self, replay_actor: ActorHandle, policy_id: PolicyID
    ) -> Dict[str, TensorType]:
        """Samples a batch from given replay actor and performs an update.

        Args:
            replay_actor: The replay buffer actor to sample from.
            policy_id: The ID of this policy.

        Returns:
            Dictionary of extra metadata from `compute_gradients()`.
        """
        # Sample a batch from the given replay actor.
        # Note that for better performance (less data sent through the
        # network), this policy should be co-located on the same node
        # as `replay_actor`. Such a co-location step is usually done during
        # the Algorithm's `setup()` phase.
        batch = ray.get(replay_actor.replay.remote(policy_id=policy_id))
        if batch is None:
            return {}

        # Send to own learn_on_batch method for updating.
        # TODO: hack w/ `hasattr`
        if hasattr(self, "devices") and len(self.devices) > 1:
            self.load_batch_into_buffer(batch, buffer_index=0)
            return self.learn_on_loaded_batch(offset=0, buffer_index=0)
        else:
            return self.learn_on_batch(batch)

    @DeveloperAPI
    def load_batch_into_buffer(self, batch: SampleBatch, buffer_index: int = 0) -> int:
        """Bulk-loads the given SampleBatch into the devices' memories.

        The data is split equally across all the Policy's devices.
        If the data is not evenly divisible by the batch size, excess data
        should be discarded.

        Args:
            batch: The SampleBatch to load.
            buffer_index: The index of the buffer (a MultiGPUTowerStack) to use
                on the devices. The number of buffers on each device depends
                on the value of the `num_multi_gpu_tower_stacks` config key.

        Returns:
            The number of tuples loaded per device.
        """
        raise NotImplementedError

    @DeveloperAPI
    def get_num_samples_loaded_into_buffer(self, buffer_index: int = 0) -> int:
        """Returns the number of currently loaded samples in the given buffer.

        Args:
            buffer_index: The index of the buffer (a MultiGPUTowerStack)
                to use on the devices. The number of buffers on each device
                depends on the value of the `num_multi_gpu_tower_stacks` config
                key.

        Returns:
            The number of tuples loaded per device.
        """
        raise NotImplementedError

    @DeveloperAPI
    def learn_on_loaded_batch(self, offset: int = 0, buffer_index: int = 0):
        """Runs a single step of SGD on an already loaded data in a buffer.

        Runs an SGD step over a slice of the pre-loaded batch, offset by
        the `offset` argument (useful for performing n minibatch SGD
        updates repeatedly on the same, already pre-loaded data).

        Updates the model weights based on the averaged per-device gradients.

        Args:
            offset: Offset into the preloaded data. Used for pre-loading
                a train-batch once to a device, then iterating over
                (subsampling through) this batch n times doing minibatch SGD.
            buffer_index: The index of the buffer (a MultiGPUTowerStack)
                to take the already pre-loaded data from. The number of buffers
                on each device depends on the value of the
                `num_multi_gpu_tower_stacks` config key.

        Returns:
            The outputs of extra_ops evaluated over the batch.
        """
        raise NotImplementedError

    @DeveloperAPI
    def compute_gradients(
        self, postprocessed_batch: SampleBatch
    ) -> Tuple[ModelGradients, Dict[str, TensorType]]:
        """Computes gradients given a batch of experiences.

        Either this in combination with `apply_gradients()` or
        `learn_on_batch()` must be implemented by subclasses.

        Args:
            postprocessed_batch: The SampleBatch object to use
                for calculating gradients.

        Returns:
            grads: List of gradient output values.
            grad_info: Extra policy-specific info values.
        """
        raise NotImplementedError

    @DeveloperAPI
    def apply_gradients(self, gradients: ModelGradients) -> None:
        """Applies the (previously) computed gradients.

        Either this in combination with `compute_gradients()` or
        `learn_on_batch()` must be implemented by subclasses.

        Args:
            gradients: The already calculated gradients to apply to this
                Policy.
        """
        raise NotImplementedError

    @DeveloperAPI
    def get_weights(self) -> ModelWeights:
        """Returns model weights.

        Note: The return value of this method will reside under the "weights"
        key in the return value of Policy.get_state(). Model weights are only
        one part of a Policy's state. Other state information contains:
        optimizer variables, exploration state, and global state vars such as
        the sampling timestep.

        Returns:
            Serializable copy or view of model weights.
        """
        raise NotImplementedError

    @DeveloperAPI
    def set_weights(self, weights: ModelWeights) -> None:
        """Sets this Policy's model's weights.

        Note: Model weights are only one part of a Policy's state. Other
        state information contains: optimizer variables, exploration state,
        and global state vars such as the sampling timestep.

        Args:
            weights: Serializable copy or view of model weights.
        """
        raise NotImplementedError

    @DeveloperAPI
    def get_exploration_state(self) -> Dict[str, TensorType]:
        """Returns the state of this Policy's exploration component.

        Returns:
            Serializable information on the `self.exploration` object.
        """
        return self.exploration.get_state()

    @DeveloperAPI
    def is_recurrent(self) -> bool:
        """Whether this Policy holds a recurrent Model.

        Returns:
            True if this Policy has-a RNN-based Model.
        """
        return False

    @DeveloperAPI
    def num_state_tensors(self) -> int:
        """The number of internal states needed by the RNN-Model of the Policy.

        Returns:
            int: The number of RNN internal states kept by this Policy's Model.
        """
        return 0

    @DeveloperAPI
    def get_initial_state(self) -> List[TensorType]:
        """Returns initial RNN state for the current policy.

        Returns:
            List[TensorType]: Initial RNN state for the current policy.
        """
        return []

    @DeveloperAPI
    @OverrideToImplementCustomLogic_CallToSuperRecommended
    def get_state(self) -> PolicyState:
        """Returns the entire current state of this Policy.

        Note: Not to be confused with an RNN model's internal state.
        State includes the Model(s)' weights, optimizer weights,
        the exploration component's state, as well as global variables, such
        as sampling timesteps.

        Note that the state may contain references to the original variables.
        This means that you may need to deepcopy() the state before mutating it.

        Returns:
            Serialized local state.
        """
        state = {
            # All the policy's weights.
            "weights": self.get_weights(),
            # The current global timestep.
            "global_timestep": self.global_timestep,
            # The current num_grad_updates counter.
            "num_grad_updates": self.num_grad_updates,
        }

        # Add this Policy's spec so it can be retreived w/o access to the original
        # code.
        policy_spec = PolicySpec(
            policy_class=type(self),
            observation_space=self.observation_space,
            action_space=self.action_space,
            config=self.config,
        )
        state["policy_spec"] = policy_spec.serialize()

        if self.config.get("enable_connectors", False):
            # Checkpoint connectors state as well if enabled.
            connector_configs = {}
            if self.agent_connectors:
                connector_configs["agent"] = self.agent_connectors.to_state()
            if self.action_connectors:
                connector_configs["action"] = self.action_connectors.to_state()
            state["connector_configs"] = connector_configs

        return state

    @PublicAPI(stability="alpha")
    def restore_connectors(self, state: PolicyState):
        """Restore agent and action connectors if configs available.

        Args:
            state: The new state to set this policy to. Can be
                obtained by calling `self.get_state()`.
        """
        # To avoid a circular dependency problem cause by SampleBatch.
        from ray.rllib.connectors.util import restore_connectors_for_policy

        # No-op if connector is not enabled.
        if not self.config.get("enable_connectors", False):
            return

        connector_configs = state.get("connector_configs", {})
        if "agent" in connector_configs:
            self.agent_connectors = restore_connectors_for_policy(
                self, connector_configs["agent"]
            )
            logger.debug("restoring agent connectors:")
            logger.debug(self.agent_connectors.__str__(indentation=4))
        if "action" in connector_configs:
            self.action_connectors = restore_connectors_for_policy(
                self, connector_configs["action"]
            )
            logger.debug("restoring action connectors:")
            logger.debug(self.action_connectors.__str__(indentation=4))

    @DeveloperAPI
    @OverrideToImplementCustomLogic_CallToSuperRecommended
    def set_state(self, state: PolicyState) -> None:
        """Restores the entire current state of this Policy from `state`.

        Args:
            state: The new state to set this policy to. Can be
                obtained by calling `self.get_state()`.
        """
        if "policy_spec" in state:
            policy_spec = PolicySpec.deserialize(state["policy_spec"])
            # Assert spaces remained the same.
            if (
                policy_spec.observation_space is not None
                and policy_spec.observation_space != self.observation_space
            ):
                logger.warning(
                    "`observation_space` in given policy state ("
                    f"{policy_spec.observation_space}) does not match this Policy's "
                    f"observation space ({self.observation_space})."
                )
            if (
                policy_spec.action_space is not None
                and policy_spec.action_space != self.action_space
            ):
                logger.warning(
                    "`action_space` in given policy state ("
                    f"{policy_spec.action_space}) does not match this Policy's "
                    f"action space ({self.action_space})."
                )
            # Override config, if part of the spec.
            if policy_spec.config:
                self.config = policy_spec.config

        # Override NN weights.
        self.set_weights(state["weights"])
        self.restore_connectors(state)

    @ExperimentalAPI
    def apply(
        self,
        func: Callable[["Policy", Optional[Any], Optional[Any]], T],
        *args,
        **kwargs,
    ) -> T:
        """Calls the given function with this Policy instance.

        Useful for when the Policy class has been converted into a ActorHandle
        and the user needs to execute some functionality (e.g. add a property)
        on the underlying policy object.

        Args:
            func: The function to call, with this Policy as first
                argument, followed by args, and kwargs.
            args: Optional additional args to pass to the function call.
            kwargs: Optional additional kwargs to pass to the function call.

        Returns:
            The return value of the function call.
        """
        return func(self, *args, **kwargs)

    @DeveloperAPI
    def on_global_var_update(self, global_vars: Dict[str, TensorType]) -> None:
        """Called on an update to global vars.

        Args:
            global_vars: Global variables by str key, broadcast from the
                driver.
        """
        # Store the current global time step (sum over all policies' sample
        # steps).
        # Make sure, we keep global_timestep as a Tensor for tf-eager
        # (leads to memory leaks if not doing so).
        if self.framework == "tf2":
            self.global_timestep.assign(global_vars["timestep"])
        else:
            self.global_timestep = global_vars["timestep"]
        # Update our lifetime gradient update counter.
        num_grad_updates = global_vars.get("num_grad_updates")
        if num_grad_updates is not None:
            self.num_grad_updates = num_grad_updates

    @DeveloperAPI
    def export_checkpoint(
        self,
        export_dir: str,
        filename_prefix=DEPRECATED_VALUE,
        *,
        policy_state: Optional[PolicyState] = None,
        checkpoint_format: str = "cloudpickle",
    ) -> None:
        """Exports Policy checkpoint to a local directory and returns an AIR Checkpoint.

        Args:
            export_dir: Local writable directory to store the AIR Checkpoint
                information into.
            policy_state: An optional PolicyState to write to disk. Used by
                `Algorithm.save_checkpoint()` to save on the additional
                `self.get_state()` calls of its different Policies.
            checkpoint_format: Either one of 'cloudpickle' or 'msgpack'.

        Example:
            >>> from ray.rllib.algorithms.ppo import PPOTorchPolicy
            >>> policy = PPOTorchPolicy(...) # doctest: +SKIP
            >>> policy.export_checkpoint("/tmp/export_dir") # doctest: +SKIP
        """
        # `filename_prefix` should not longer be used as new Policy checkpoints
        # contain more than one file with a fixed filename structure.
        if filename_prefix != DEPRECATED_VALUE:
            deprecation_warning(
                old="Policy.export_checkpoint(filename_prefix=...)",
                error=True,
            )
        if checkpoint_format not in ["cloudpickle", "msgpack"]:
            raise ValueError(
                f"Value of `checkpoint_format` ({checkpoint_format}) must either be "
                "'cloudpickle' or 'msgpack'!"
            )

        if policy_state is None:
            policy_state = self.get_state()

        # Write main policy state file.
        os.makedirs(export_dir, exist_ok=True)
        if checkpoint_format == "cloudpickle":
            policy_state["checkpoint_version"] = CHECKPOINT_VERSION
            state_file = "policy_state.pkl"
            with open(os.path.join(export_dir, state_file), "w+b") as f:
                pickle.dump(policy_state, f)
        else:
            from ray.rllib.algorithms.algorithm_config import AlgorithmConfig

            msgpack = try_import_msgpack(error=True)
            policy_state["checkpoint_version"] = str(CHECKPOINT_VERSION)
            # Serialize the config for msgpack dump'ing.
            policy_state["policy_spec"]["config"] = AlgorithmConfig._serialize_dict(
                policy_state["policy_spec"]["config"]
            )
            state_file = "policy_state.msgpck"
            with open(os.path.join(export_dir, state_file), "w+b") as f:
                msgpack.dump(policy_state, f)

        # Write RLlib checkpoint json.
        with open(os.path.join(export_dir, "rllib_checkpoint.json"), "w") as f:
            json.dump(
                {
                    "type": "Policy",
                    "checkpoint_version": str(policy_state["checkpoint_version"]),
                    "format": checkpoint_format,
                    "state_file": state_file,
                    "ray_version": ray.__version__,
                    "ray_commit": ray.__commit__,
                },
                f,
            )

        # Add external model files, if required.
        if self.config["export_native_model_files"]:
            self.export_model(os.path.join(export_dir, "model"))

    @DeveloperAPI
    def export_model(self, export_dir: str, onnx: Optional[int] = None) -> None:
        """Exports the Policy's Model to local directory for serving.

        Note: The file format will depend on the deep learning framework used.
        See the child classed of Policy and their `export_model`
        implementations for more details.

        Args:
            export_dir: Local writable directory.
            onnx: If given, will export model in ONNX format. The
                value of this parameter set the ONNX OpSet version to use.

        Raises:
            ValueError: If a native DL-framework based model (e.g. a keras Model)
            cannot be saved to disk for various reasons.
        """
        raise NotImplementedError

    @DeveloperAPI
    def import_model_from_h5(self, import_file: str) -> None:
        """Imports Policy from local file.

        Args:
            import_file: Local readable file.
        """
        raise NotImplementedError

    @DeveloperAPI
    def get_session(self) -> Optional["tf1.Session"]:
        """Returns tf.Session object to use for computing actions or None.

        Note: This method only applies to TFPolicy sub-classes. All other
        sub-classes should expect a None to be returned from this method.

        Returns:
            The tf Session to use for computing actions and losses with
                this policy or None.
        """
        return None

    def get_host(self) -> str:
        """Returns the computer's network name.

        Returns:
            The computer's networks name or an empty string, if the network
            name could not be determined.
        """
        return platform.node()

    def _get_num_gpus_for_policy(self) -> int:
        """Decide on the number of CPU/GPU nodes this policy should run on.

        Return:
            0 if policy should run on CPU. >0 if policy should run on 1 or
            more GPUs.
        """
        worker_idx = self.config.get("worker_index", 0)
        fake_gpus = self.config.get("_fake_gpus", False)

        if (
            ray._private.worker._mode() == ray._private.worker.LOCAL_MODE
            and not fake_gpus
        ):
            # If in local debugging mode, and _fake_gpus is not on.
            num_gpus = 0
        elif worker_idx == 0:
            # If we are on the new RLModule/Learner stack, `num_gpus` is deprecated.
            # so use `num_gpus_per_worker` for policy sampling
            # we need this .get() syntax here to ensure backwards compatibility.
            if self.config.get("_enable_learner_api", False):
                num_gpus = self.config["num_gpus_per_worker"]
            else:
                # If head node, take num_gpus.
                num_gpus = self.config["num_gpus"]
        else:
            # If worker node, take num_gpus_per_worker
            num_gpus = self.config["num_gpus_per_worker"]

        if num_gpus == 0:
            dev = "CPU"
        else:
            dev = "{} {}".format(num_gpus, "fake-GPUs" if fake_gpus else "GPUs")

        logger.info(
            "Policy (worker={}) running on {}.".format(
                worker_idx if worker_idx > 0 else "local", dev
            )
        )

        return num_gpus

    def _create_exploration(self) -> Exploration:
        """Creates the Policy's Exploration object.

        This method only exists b/c some Algorithms do not use TfPolicy nor
        TorchPolicy, but inherit directly from Policy. Others inherit from
        TfPolicy w/o using DynamicTFPolicy.
        TODO(sven): unify these cases.

        Returns:
            Exploration: The Exploration object to be used by this Policy.
        """
        if getattr(self, "exploration", None) is not None:
            return self.exploration

        exploration = from_config(
            Exploration,
            self.config.get("exploration_config", {"type": "StochasticSampling"}),
            action_space=self.action_space,
            policy_config=self.config,
            model=getattr(self, "model", None),
            num_workers=self.config.get("num_workers", 0),
            worker_index=self.config.get("worker_index", 0),
            framework=getattr(self, "framework", self.config.get("framework", "tf")),
        )
        return exploration

    def _get_default_view_requirements(self):
        """Returns a default ViewRequirements dict.

        Note: This is the base/maximum requirement dict, from which later
        some requirements will be subtracted again automatically to streamline
        data collection, batch creation, and data transfer.

        Returns:
            ViewReqDict: The default view requirements dict.
        """

        # Default view requirements (equal to those that we would use before
        # the trajectory view API was introduced).
        return {
            SampleBatch.OBS: ViewRequirement(space=self.observation_space),
            SampleBatch.NEXT_OBS: ViewRequirement(
                data_col=SampleBatch.OBS,
                shift=1,
                space=self.observation_space,
                used_for_compute_actions=False,
            ),
            SampleBatch.ACTIONS: ViewRequirement(
                space=self.action_space, used_for_compute_actions=False
            ),
            # For backward compatibility with custom Models that don't specify
            # these explicitly (will be removed by Policy if not used).
            SampleBatch.PREV_ACTIONS: ViewRequirement(
                data_col=SampleBatch.ACTIONS, shift=-1, space=self.action_space
            ),
            SampleBatch.REWARDS: ViewRequirement(),
            # For backward compatibility with custom Models that don't specify
            # these explicitly (will be removed by Policy if not used).
            SampleBatch.PREV_REWARDS: ViewRequirement(
                data_col=SampleBatch.REWARDS, shift=-1
            ),
            SampleBatch.TERMINATEDS: ViewRequirement(),
            SampleBatch.TRUNCATEDS: ViewRequirement(),
            SampleBatch.INFOS: ViewRequirement(used_for_compute_actions=False),
            SampleBatch.EPS_ID: ViewRequirement(),
            SampleBatch.UNROLL_ID: ViewRequirement(),
            SampleBatch.AGENT_INDEX: ViewRequirement(),
            SampleBatch.T: ViewRequirement(),
        }

    def _initialize_loss_from_dummy_batch(
        self,
        auto_remove_unneeded_view_reqs: bool = True,
        stats_fn=None,
    ) -> None:
        """Performs test calls through policy's model and loss.

        NOTE: This base method should work for define-by-run Policies such as
        torch and tf-eager policies.

        If required, will thereby detect automatically, which data views are
        required by a) the forward pass, b) the postprocessing, and c) the loss
        functions, and remove those from self.view_requirements that are not
        necessary for these computations (to save data storage and transfer).

        Args:
            auto_remove_unneeded_view_reqs: Whether to automatically
                remove those ViewRequirements records from
                self.view_requirements that are not needed.
            stats_fn (Optional[Callable[[Policy, SampleBatch], Dict[str,
                TensorType]]]): An optional stats function to be called after
                the loss.
        """

        if self.config.get("_disable_initialize_loss_from_dummy_batch", False):
            return
        # Signal Policy that currently we do not like to eager/jit trace
        # any function calls. This is to be able to track, which columns
        # in the dummy batch are accessed by the different function (e.g.
        # loss) such that we can then adjust our view requirements.
        self._no_tracing = True
        # Save for later so that loss init does not change global timestep
        global_ts_before_init = int(convert_to_numpy(self.global_timestep))

        sample_batch_size = min(
            max(self.batch_divisibility_req * 4, 32),
            self.config["train_batch_size"],  # Don't go over the asked batch size.
        )
        self._dummy_batch = self._get_dummy_batch_from_view_requirements(
            sample_batch_size
        )
        self._lazy_tensor_dict(self._dummy_batch)
        # With RL Modules you want the explore flag to be True for initialization
        # of the tensors and placeholder you'd need for training.
        explore = self.config.get("_enable_rl_module_api", False)

        actions, state_outs, extra_outs = self.compute_actions_from_input_dict(
            self._dummy_batch, explore=explore
        )
        if not self.config.get("_enable_rl_module_api", False):
            for key, view_req in self.view_requirements.items():
                if key not in self._dummy_batch.accessed_keys:
                    view_req.used_for_compute_actions = False
        # Add all extra action outputs to view reqirements (these may be
        # filtered out later again, if not needed for postprocessing or loss).
        for key, value in extra_outs.items():
            self._dummy_batch[key] = value
            if key not in self.view_requirements:
                if isinstance(value, (dict, np.ndarray)):
                    # the assumption is that value is a nested_dict of np.arrays leaves
                    space = get_gym_space_from_struct_of_tensors(value)
                    self.view_requirements[key] = ViewRequirement(
                        space=space, used_for_compute_actions=False
                    )
                else:
                    raise ValueError(
                        "policy.compute_actions_from_input_dict() returns an "
                        "extra action output that is neither a numpy array nor a dict."
                    )

        for key in self._dummy_batch.accessed_keys:
            if key not in self.view_requirements:
                self.view_requirements[key] = ViewRequirement()
                self.view_requirements[key].used_for_compute_actions = False
            # TODO (kourosh) Why did we use to make used_for_compute_actions True here?
        new_batch = self._get_dummy_batch_from_view_requirements(sample_batch_size)
        # Make sure the dummy_batch will return numpy arrays when accessed
        self._dummy_batch.set_get_interceptor(None)

        # try to re-use the output of the previous run to avoid overriding things that
        # would break (e.g. scale = 0 of Normal distribution cannot be zero)
        for k in new_batch:
            if k not in self._dummy_batch:
                self._dummy_batch[k] = new_batch[k]

        # Make sure the book-keeping of dummy_batch keys are reset to correcly track
        # what is accessed, what is added and what's deleted from now on.
        self._dummy_batch.accessed_keys.clear()
        self._dummy_batch.deleted_keys.clear()
        self._dummy_batch.added_keys.clear()

        if self.exploration:
            # Policies with RLModules don't have an exploration object.
            self.exploration.postprocess_trajectory(self, self._dummy_batch)

        postprocessed_batch = self.postprocess_trajectory(self._dummy_batch)
        seq_lens = None
        if state_outs:
            B = 4  # For RNNs, have B=4, T=[depends on sample_batch_size]
            if self.config.get("_enable_rl_module_api", False):
                sub_batch = postprocessed_batch[:B]
                postprocessed_batch["state_in"] = sub_batch["state_in"]
                postprocessed_batch["state_out"] = sub_batch["state_out"]
            else:
                i = 0
                while "state_in_{}".format(i) in postprocessed_batch:
                    postprocessed_batch["state_in_{}".format(i)] = postprocessed_batch[
                        "state_in_{}".format(i)
                    ][:B]
                    if "state_out_{}".format(i) in postprocessed_batch:
                        postprocessed_batch[
                            "state_out_{}".format(i)
                        ] = postprocessed_batch["state_out_{}".format(i)][:B]
                    i += 1

            seq_len = sample_batch_size // B
            seq_lens = np.array([seq_len for _ in range(B)], dtype=np.int32)
            postprocessed_batch[SampleBatch.SEQ_LENS] = seq_lens

        if not self.config.get("_enable_learner_api"):
            # Switch on lazy to-tensor conversion on `postprocessed_batch`.
            train_batch = self._lazy_tensor_dict(postprocessed_batch)
            # Calling loss, so set `is_training` to True.
            train_batch.set_training(True)
            if seq_lens is not None:
                train_batch[SampleBatch.SEQ_LENS] = seq_lens
            train_batch.count = self._dummy_batch.count

            # Call the loss function, if it exists.
            # TODO(jungong) : clean up after all agents get migrated.
            # We should simply do self.loss(...) here.
            if self._loss is not None:
                self._loss(self, self.model, self.dist_class, train_batch)
            elif is_overridden(self.loss) and not self.config["in_evaluation"]:
                self.loss(self.model, self.dist_class, train_batch)
            # Call the stats fn, if given.
            # TODO(jungong) : clean up after all agents get migrated.
            # We should simply do self.stats_fn(train_batch) here.
            if stats_fn is not None:
                stats_fn(self, train_batch)
            if hasattr(self, "stats_fn") and not self.config["in_evaluation"]:
                self.stats_fn(train_batch)
        else:
            # This is not needed to run a training with the Learner API, but useful if
            # we want to create a batch of data for training from view requirements.
            for key in set(postprocessed_batch.keys()).difference(
                set(new_batch.keys())
            ):
                # Add all columns generated by postprocessing to view requirements.
                if key not in self.view_requirements and key != SampleBatch.SEQ_LENS:
                    self.view_requirements[key] = ViewRequirement(
                        used_for_compute_actions=False
                    )

        # Re-enable tracing.
        self._no_tracing = False

        # Add new columns automatically to view-reqs.
        if (
            not self.config.get("_enable_learner_api")
            and auto_remove_unneeded_view_reqs
        ):
            # Add those needed for postprocessing and training.
            all_accessed_keys = (
                train_batch.accessed_keys
                | self._dummy_batch.accessed_keys
                | self._dummy_batch.added_keys
            )
            for key in all_accessed_keys:
                if key not in self.view_requirements and key != SampleBatch.SEQ_LENS:
                    self.view_requirements[key] = ViewRequirement(
                        used_for_compute_actions=False
                    )
            if self._loss or is_overridden(self.loss):
                # Tag those only needed for post-processing (with some
                # exceptions).
                for key in self._dummy_batch.accessed_keys:
                    if (
                        key not in train_batch.accessed_keys
                        and key in self.view_requirements
                        and key not in self.model.view_requirements
                        and key
                        not in [
                            SampleBatch.EPS_ID,
                            SampleBatch.AGENT_INDEX,
                            SampleBatch.UNROLL_ID,
                            SampleBatch.TERMINATEDS,
                            SampleBatch.TRUNCATEDS,
                            SampleBatch.REWARDS,
                            SampleBatch.INFOS,
                            SampleBatch.T,
                        ]
                    ):
                        self.view_requirements[key].used_for_training = False
                # Remove those not needed at all (leave those that are needed
                # by Sampler to properly execute sample collection). Also always leave
                # TERMINATEDS, TRUNCATEDS, REWARDS, INFOS, no matter what.
                for key in list(self.view_requirements.keys()):
                    if (
                        key not in all_accessed_keys
                        and key
                        not in [
                            SampleBatch.EPS_ID,
                            SampleBatch.AGENT_INDEX,
                            SampleBatch.UNROLL_ID,
                            SampleBatch.TERMINATEDS,
                            SampleBatch.TRUNCATEDS,
                            SampleBatch.REWARDS,
                            SampleBatch.INFOS,
                            SampleBatch.T,
                        ]
                        and key not in self.model.view_requirements
                    ):
                        # If user deleted this key manually in postprocessing
                        # fn, warn about it and do not remove from
                        # view-requirements.
                        if key in self._dummy_batch.deleted_keys:
                            logger.warning(
                                "SampleBatch key '{}' was deleted manually in "
                                "postprocessing function! RLlib will "
                                "automatically remove non-used items from the "
                                "data stream. Remove the `del` from your "
                                "postprocessing function.".format(key)
                            )
                        # If we are not writing output to disk, save to erase
                        # this key to save space in the sample batch.
                        elif self.config["output"] is None:
                            del self.view_requirements[key]

        if type(self.global_timestep) is int:
            self.global_timestep = global_ts_before_init
        elif isinstance(self.global_timestep, tf.Variable):
            self.global_timestep.assign(global_ts_before_init)
        else:
            raise ValueError(
                "Variable self.global_timestep of policy {} needs to be "
                "either of type `int` or `tf.Variable`, "
                "but is of type {}.".format(self, type(self.global_timestep))
            )

    @ExperimentalAPI
    def maybe_add_time_dimension(
        self,
        input_dict: Dict[str, TensorType],
        seq_lens: TensorType,
        framework: str = None,
    ):
        """Adds a time dimension for recurrent RLModules.

        Args:
            input_dict: The input dict.
            seq_lens: The sequence lengths.
            framework: The framework to use for adding the time dimensions.
                If None, will default to the framework of the policy.

        Returns:
            The input dict, with a possibly added time dimension.
        """
        # We need to check for hasattr(self, "model") because a dummy Policy may not
        # have a model.
        if (
            self.config.get("_enable_rl_module_api", False)
            and hasattr(self, "model")
            and self.model.is_stateful()
        ):
            # Note that this is a temporary workaround to fit the old sampling stack
            # to RL Modules.
            ret = {}
            framework = framework or self.model.framework

            def _add_time_dimension(inputs):
                inputs = add_time_dimension(
                    inputs,
                    seq_lens=seq_lens,
                    framework=framework,
                    time_major=self.config.get("model", {}).get("_time_major", False),
                )
                return inputs

            def _add_state_out_time_dimension(inputs):
                # We do a hack here in that we add a time dimension,
                # even though the tensor already has one. Then, we remove the
                # original time dimension.
                v_w_two_time_dims = _add_time_dimension(inputs)
                if framework == "tf2":
                    return tf.squeeze(v_w_two_time_dims, axis=2)
                elif framework == "torch":
                    # Remove second time dimensions
                    return torch.squeeze(v_w_two_time_dims, axis=2)
                elif framework == "np":
                    shape = v_w_two_time_dims.shape
                    padded_batch_dim = shape[0]
                    padded_time_dim = shape[1]
                    other_dims = shape[3:]
                    new_shape = (padded_batch_dim, padded_time_dim) + other_dims
                    return v_w_two_time_dims.reshape(new_shape)
                else:
                    raise ValueError(f"Framework {framework} not implemented!")

            for k, v in input_dict.items():
                if k == SampleBatch.INFOS:
                    ret[k] = _add_time_dimension(v)
                elif k == SampleBatch.SEQ_LENS:
                    # sequence lengths have no time dimension
                    ret[k] = v
                elif k == STATE_IN:
                    # Assume that batch_repeat_value is max seq len.
                    # This is commonly the case for STATE_IN
                    # Values should already have correct batch and time dimension
                    assert self.view_requirements[k].batch_repeat_value != 1
                    ret[k] = v
                elif k == STATE_OUT:
                    # Assume that batch_repeat_value is 1
                    # This is commonly the case for STATE_OUT
                    assert self.view_requirements[k].batch_repeat_value == 1
                    ret[k] = tree.map_structure(_add_state_out_time_dimension, v)
                else:
                    ret[k] = tree.map_structure(_add_time_dimension, v)

            return SampleBatch(ret)
        else:
            return input_dict

    @ExperimentalAPI
    def maybe_remove_time_dimension(self, input_dict: Dict[str, TensorType]):
        """Removes a time dimension for recurrent RLModules.

        Args:
            input_dict: The input dict.

        Returns:
            The input dict with a possibly removed time dimension.
        """
        raise NotImplementedError

    def _get_dummy_batch_from_view_requirements(
        self, batch_size: int = 1
    ) -> SampleBatch:
        """Creates a numpy dummy batch based on the Policy's view requirements.

        Args:
            batch_size: The size of the batch to create.

        Returns:
            Dict[str, TensorType]: The dummy batch containing all zero values.
        """
        ret = {}
        for view_col, view_req in self.view_requirements.items():
            data_col = view_req.data_col or view_col
            # Flattened dummy batch.
            if (isinstance(view_req.space, (gym.spaces.Tuple, gym.spaces.Dict))) and (
                (
                    data_col == SampleBatch.OBS
                    and not self.config["_disable_preprocessor_api"]
                )
                or (
                    data_col == SampleBatch.ACTIONS
                    and not self.config.get("_disable_action_flattening")
                )
            ):
                _, shape = ModelCatalog.get_action_shape(
                    view_req.space, framework=self.config["framework"]
                )
                ret[view_col] = np.zeros((batch_size,) + shape[1:], np.float32)
            # Non-flattened dummy batch.
            else:
                # Range of indices on time-axis, e.g. "-50:-1".
                if isinstance(view_req.space, gym.spaces.Space):
                    time_size = (
                        len(view_req.shift_arr) if len(view_req.shift_arr) > 1 else None
                    )
                    ret[view_col] = get_dummy_batch_for_space(
                        view_req.space, batch_size=batch_size, time_size=time_size
                    )
                else:
                    ret[view_col] = [view_req.space for _ in range(batch_size)]

        # Due to different view requirements for the different columns,
        # columns in the resulting batch may not all have the same batch size.
        return SampleBatch(ret)

    def _update_model_view_requirements_from_init_state(self):
        """Uses Model's (or this Policy's) init state to add needed ViewReqs.

        Can be called from within a Policy to make sure RNNs automatically
        update their internal state-related view requirements.
        Changes the `self.view_requirements` dict.
        """
        self._model_init_state_automatically_added = True
        model = getattr(self, "model", None)

        obj = model or self
        if model and not hasattr(model, "view_requirements"):
            model.view_requirements = {
                SampleBatch.OBS: ViewRequirement(space=self.observation_space)
            }
        view_reqs = obj.view_requirements
        # Add state-ins to this model's view.
        init_state = []
        if hasattr(obj, "get_initial_state") and callable(obj.get_initial_state):
            init_state = obj.get_initial_state()
        else:
            # Add this functionality automatically for new native model API.
            if (
                tf
                and isinstance(model, tf.keras.Model)
                and "state_in_0" not in view_reqs
            ):
                obj.get_initial_state = lambda: [
                    np.zeros_like(view_req.space.sample())
                    for k, view_req in model.view_requirements.items()
                    if k.startswith("state_in_")
                ]
            else:
                obj.get_initial_state = lambda: []
                if "state_in_0" in view_reqs:
                    self.is_recurrent = lambda: True

        # Make sure auto-generated init-state view requirements get added
        # to both Policy and Model, no matter what.
        view_reqs = [view_reqs] + (
            [self.view_requirements] if hasattr(self, "view_requirements") else []
        )

        for i, state in enumerate(init_state):
            # Allow `state` to be either a Space (use zeros as initial values)
            # or any value (e.g. a dict or a non-zero tensor).
            fw = (
                np
                if isinstance(state, np.ndarray)
                else torch
                if torch and torch.is_tensor(state)
                else None
            )
            if fw:
                space = (
                    Box(-1.0, 1.0, shape=state.shape) if fw.all(state == 0.0) else state
                )
            else:
                space = state
            for vr in view_reqs:
                # Only override if user has not already provided
                # custom view-requirements for state_in_n.
                if "state_in_{}".format(i) not in vr:
                    vr["state_in_{}".format(i)] = ViewRequirement(
                        "state_out_{}".format(i),
                        shift=-1,
                        used_for_compute_actions=True,
                        batch_repeat_value=self.config.get("model", {}).get(
                            "max_seq_len", 1
                        ),
                        space=space,
                    )
                # Only override if user has not already provided
                # custom view-requirements for state_out_n.
                if "state_out_{}".format(i) not in vr:
                    vr["state_out_{}".format(i)] = ViewRequirement(
                        space=space, used_for_training=True
                    )

    @DeveloperAPI
    def __repr__(self):
        return type(self).__name__

    @Deprecated(new="get_exploration_state", error=True)
    def get_exploration_info(self) -> Dict[str, TensorType]:
        return self.get_exploration_state()


@DeveloperAPI
def get_gym_space_from_struct_of_tensors(
    value: Union[Mapping, Tuple, List, TensorType],
    batched_input=True,
) -> gym.Space:

    start_idx = 1 if batched_input else 0
    struct = tree.map_structure(
        lambda x: gym.spaces.Box(
            -1.0, 1.0, shape=x.shape[start_idx:], dtype=get_np_dtype(x)
        ),
        value,
    )
    space = get_gym_space_from_struct_of_spaces(struct)
    return space


@DeveloperAPI
def get_gym_space_from_struct_of_spaces(value: Union[Dict, Tuple]) -> gym.spaces.Dict:
    if isinstance(value, Mapping):
        return gym.spaces.Dict(
            {k: get_gym_space_from_struct_of_spaces(v) for k, v in value.items()}
        )
    elif isinstance(value, (tuple, list)):
        return gym.spaces.Tuple([get_gym_space_from_struct_of_spaces(v) for v in value])
    else:
        assert isinstance(value, gym.spaces.Space), (
            f"The struct of spaces should only contain dicts, tiples and primitive "
            f"gym spaces. Space is of type {type(value)}"
        )
        return value



import base64
import collections
import errno
import io
import json
import logging
import mmap
import multiprocessing
import os
import random
import shutil
import signal
import socket
import subprocess
import sys
import time
from pathlib import Path
from typing import List, Optional, IO, AnyStr

# Import psutil after ray so the packaged version is used.
import psutil

# Ray modules
import ray
import ray._private.ray_constants as ray_constants
from ray._raylet import GcsClient, GcsClientOptions
from ray.core.generated.common_pb2 import Language

resource = None
if sys.platform != "win32":
    _timeout = 30
else:
    _timeout = 60

EXE_SUFFIX = ".exe" if sys.platform == "win32" else ""

# True if processes are run in the valgrind profiler.
RUN_RAYLET_PROFILER = False

# Location of the redis server.
RAY_HOME = os.path.join(os.path.dirname(os.path.dirname(__file__)), "..", "..")
RAY_PATH = os.path.abspath(os.path.dirname(os.path.dirname(__file__)))
RAY_PRIVATE_DIR = "_private"
AUTOSCALER_PRIVATE_DIR = os.path.join("autoscaler", "_private")

# Location of the raylet executables.
RAYLET_EXECUTABLE = os.path.join(
    RAY_PATH, "core", "src", "ray", "raylet", "raylet" + EXE_SUFFIX
)
GCS_SERVER_EXECUTABLE = os.path.join(
    RAY_PATH, "core", "src", "ray", "gcs", "gcs_server" + EXE_SUFFIX
)

# Location of the cpp default worker executables.
DEFAULT_WORKER_EXECUTABLE = os.path.join(RAY_PATH, "cpp", "default_worker" + EXE_SUFFIX)

# Location of the native libraries.
DEFAULT_NATIVE_LIBRARY_PATH = os.path.join(RAY_PATH, "cpp", "lib")

DASHBOARD_DEPENDENCY_ERROR_MESSAGE = (
    "Not all Ray Dashboard dependencies were "
    "found. To use the dashboard please "
    "install Ray using `pip install "
    "ray[default]`."
)

RAY_JEMALLOC_LIB_PATH = "RAY_JEMALLOC_LIB_PATH"
RAY_JEMALLOC_CONF = "RAY_JEMALLOC_CONF"
RAY_JEMALLOC_PROFILE = "RAY_JEMALLOC_PROFILE"

# Comma separated name of components that will run memory profiler.
# Ray uses `memray` to memory profile internal components.
# The name of the component must be one of ray_constants.PROCESS_TYPE*.
RAY_MEMRAY_PROFILE_COMPONENT_ENV = "RAY_INTERNAL_MEM_PROFILE_COMPONENTS"
# Options to specify for `memray run` command. See
# `memray run --help` for more details.
# Example:
# RAY_INTERNAL_MEM_PROFILE_OPTIONS="--live,--live-port,3456,-q,"
# -> `memray run --live --live-port 3456 -q`
RAY_MEMRAY_PROFILE_OPTIONS_ENV = "RAY_INTERNAL_MEM_PROFILE_OPTIONS"

# Logger for this module. It should be configured at the entry point
# into the program using Ray. Ray provides a default configuration at
# entry/init points.
logger = logging.getLogger(__name__)

ProcessInfo = collections.namedtuple(
    "ProcessInfo",
    [
        "process",
        "stdout_file",
        "stderr_file",
        "use_valgrind",
        "use_gdb",
        "use_valgrind_profiler",
        "use_perftools_profiler",
        "use_tmux",
    ],
)


def _site_flags() -> List[str]:
    """Detect whether flags related to site packages are enabled for the current
    interpreter. To run Ray in hermetic build environments, it helps to pass these flags
    down to Python workers.
    """
    flags = []
    # sys.flags hidden behind helper methods for unit testing.
    if _no_site():
        flags.append("-S")
    if _no_user_site():
        flags.append("-s")
    return flags


# sys.flags hidden behind helper methods for unit testing.
def _no_site():
    return sys.flags.no_site


# sys.flags hidden behind helper methods for unit testing.
def _no_user_site():
    return sys.flags.no_user_site


def _build_python_executable_command_memory_profileable(
    component: str, session_dir: str, unbuffered: bool = True
):
    """Build the Python executable command.

    It runs a memory profiler if env var is configured.

    Args:
        component: Name of the component. It must be one of
            ray_constants.PROCESS_TYPE*.
        session_dir: The directory name of the Ray session.
        unbuffered: If true, Python executable is started with unbuffered option.
            e.g., `-u`.
            It means the logs are flushed immediately (good when there's a failure),
            but writing to a log file can be slower.
    """
    command = [
        sys.executable,
    ]
    if unbuffered:
        command.append("-u")
    components_to_memory_profile = os.getenv(RAY_MEMRAY_PROFILE_COMPONENT_ENV, "")
    if not components_to_memory_profile:
        return command

    components_to_memory_profile = set(components_to_memory_profile.split(","))
    try:
        import memray  # noqa: F401
    except ImportError:
        raise ImportError(
            "Memray is required to memory profiler on components "
            f"{components_to_memory_profile}. Run `pip install memray`."
        )
    if component in components_to_memory_profile:
        session_dir = Path(session_dir)
        session_name = session_dir.name
        profile_dir = session_dir / "profile"
        profile_dir.mkdir(exist_ok=True)
        output_file_path = profile_dir / f"{session_name}_memory_{component}.bin"
        options = os.getenv(RAY_MEMRAY_PROFILE_OPTIONS_ENV, None)
        options = options.split(",") if options else []
        command.extend(["-m", "memray", "run", "-o", str(output_file_path), *options])

    return command


def _get_gcs_client_options(gcs_server_address):
    return GcsClientOptions.from_gcs_address(gcs_server_address)


def serialize_config(config):
    return base64.b64encode(json.dumps(config).encode("utf-8")).decode("utf-8")


def propagate_jemalloc_env_var(
    *,
    jemalloc_path: str,
    jemalloc_conf: str,
    jemalloc_comps: List[str],
    process_type: str,
):
    """Read the jemalloc memory profiling related
    env var and return the dictionary that translates
    them to proper jemalloc related env vars.

    For example, if users specify `RAY_JEMALLOC_LIB_PATH`,
    it is translated into `LD_PRELOAD` which is needed to
    run Jemalloc as a shared library.

    Params:
        jemalloc_path: The path to the jemalloc shared library.
        jemalloc_conf: `,` separated string of jemalloc config.
        jemalloc_comps: The list of Ray components
            that we will profile.
        process_type: The process type that needs jemalloc
            env var for memory profiling. If it doesn't match one of
            jemalloc_comps, the function will return an empty dict.

    Returns:
        dictionary of {env_var: value}
            that are needed to jemalloc profiling. The caller can
            call `dict.update(return_value_of_this_func)` to
            update the dict of env vars. If the process_type doesn't
            match jemalloc_comps, it will return an empty dict.
    """
    assert isinstance(jemalloc_comps, list)
    assert process_type is not None
    process_type = process_type.lower()
    if not jemalloc_path or process_type not in jemalloc_comps:
        return {}

    env_vars = {
        "LD_PRELOAD": jemalloc_path,
    }
    if jemalloc_conf:
        env_vars.update({"MALLOC_CONF": jemalloc_conf})
    return env_vars


class ConsolePopen(subprocess.Popen):
    if sys.platform == "win32":

        def terminate(self):
            if isinstance(self.stdin, io.IOBase):
                self.stdin.close()
            if self._use_signals:
                self.send_signal(signal.CTRL_BREAK_EVENT)
            else:
                super(ConsolePopen, self).terminate()

        def __init__(self, *args, **kwargs):
            # CREATE_NEW_PROCESS_GROUP is used to send Ctrl+C on Windows:
            # https://docs.python.org/3/library/subprocess.html#subprocess.Popen.send_signal
            new_pgroup = subprocess.CREATE_NEW_PROCESS_GROUP
            flags_to_add = 0
            if ray._private.utils.detect_fate_sharing_support():
                # If we don't have kernel-mode fate-sharing, then don't do this
                # because our children need to be in out process group for
                # the process reaper to properly terminate them.
                flags_to_add = new_pgroup
            flags_key = "creationflags"
            if flags_to_add:
                kwargs[flags_key] = (kwargs.get(flags_key) or 0) | flags_to_add
            self._use_signals = kwargs[flags_key] & new_pgroup
            super(ConsolePopen, self).__init__(*args, **kwargs)


def address(ip_address, port):
    return ip_address + ":" + str(port)


def new_port(lower_bound=10000, upper_bound=65535, denylist=None):
    if not denylist:
        denylist = set()
    port = random.randint(lower_bound, upper_bound)
    retry = 0
    while port in denylist:
        if retry > 100:
            break
        port = random.randint(lower_bound, upper_bound)
        retry += 1
    if retry > 100:
        raise ValueError(
            "Failed to find a new port from the range "
            f"{lower_bound}-{upper_bound}. Denylist: {denylist}"
        )
    return port


def _find_address_from_flag(flag: str):
    """
    Attempts to find all valid Ray addresses on this node, specified by the
    flag.

    Params:
        flag: `--redis-address` or `--gcs-address`
    Returns:
        Set of detected addresses.
    """
    # Using Redis address `--redis-address` as an example:
    # Currently, this extracts the deprecated --redis-address from the command
    # that launched the raylet running on this node, if any. Anyone looking to
    # edit this function should be warned that these commands look like, for
    # example:
    # /usr/local/lib/python3.8/dist-packages/ray/core/src/ray/raylet/raylet
    # --redis_address=123.456.78.910 --node_ip_address=123.456.78.910
    # --raylet_socket_name=... --store_socket_name=... --object_manager_port=0
    # --min_worker_port=10000 --max_worker_port=19999
    # --node_manager_port=58578 --redis_port=6379
    # --maximum_startup_concurrency=8
    # --static_resource_list=node:123.456.78.910,1.0,object_store_memory,66
    # --config_list=plasma_store_as_thread,True
    # --python_worker_command=/usr/bin/python
    #     /usr/local/lib/python3.8/dist-packages/ray/workers/default_worker.py
    #     --redis-address=123.456.78.910:6379
    #     --node-ip-address=123.456.78.910 --node-manager-port=58578
    #     --object-store-name=... --raylet-name=...
    #     --temp-dir=/tmp/ray
    #     --metrics-agent-port=41856 --redis-password=[MASKED]
    #     --java_worker_command= --cpp_worker_command=
    #     --redis_password=[MASKED] --temp_dir=/tmp/ray --session_dir=...
    #     --metrics-agent-port=41856 --metrics_export_port=64229
    #     --agent_command=/usr/bin/python
    #     -u /usr/local/lib/python3.8/dist-packages/ray/dashboard/agent.py
    #         --redis-address=123.456.78.910:6379 --metrics-export-port=64229
    #         --dashboard-agent-port=41856 --node-manager-port=58578
    #         --object-store-name=... --raylet-name=... --temp-dir=/tmp/ray
    #         --log-dir=/tmp/ray/session_2020-11-08_14-29-07_199128_278000/logs
    #         --redis-password=[MASKED] --object_store_memory=5037192806
    #         --plasma_directory=/tmp
    # Longer arguments are elided with ... but all arguments from this instance
    # are included, to provide a sense of what is in these.
    # Indeed, we had to pull --redis-address to the front of each call to make
    # this readable.
    # As you can see, this is very long and complex, which is why we can't
    # simply extract all the the arguments using regular expressions and
    # present a dict as if we never lost track of these arguments, for
    # example. Picking out --redis-address below looks like it might grab the
    # wrong thing, but double-checking that we're finding the correct process
    # by checking that the contents look like we expect would probably be prone
    # to choking in unexpected ways.
    # Notice that --redis-address appears twice. This is not a copy-paste
    # error; this is the reason why the for loop below attempts to pick out
    # every appearance of --redis-address.

    # The --redis-address here is what is now called the --address, but it
    # appears in the default_worker.py and agent.py calls as --redis-address.
    pids = psutil.pids()
    addresses = set()
    for pid in pids:
        try:
            proc = psutil.Process(pid)
            # HACK: Workaround for UNIX idiosyncrasy
            # Normally, cmdline() is supposed to return the argument list.
            # But it in some cases (such as when setproctitle is called),
            # an arbitrary string resembling a command-line is stored in
            # the first argument.
            # Explanation: https://unix.stackexchange.com/a/432681
            # More info: https://github.com/giampaolo/psutil/issues/1179
            cmdline = proc.cmdline()
            # NOTE(kfstorm): To support Windows, we can't use
            # `os.path.basename(cmdline[0]) == "raylet"` here.

            if len(cmdline) > 0 and "raylet" in os.path.basename(cmdline[0]):
                for arglist in cmdline:
                    # Given we're merely seeking --redis-address, we just split
                    # every argument on spaces for now.
                    for arg in arglist.split(" "):
                        # TODO(ekl): Find a robust solution for locating Redis.
                        if arg.startswith(flag):
                            proc_addr = arg.split("=")[1]
                            # TODO(mwtian): remove this workaround after Ray
                            # no longer sets --redis-address to None.
                            if proc_addr != "" and proc_addr != "None":
                                addresses.add(proc_addr)
        except psutil.AccessDenied:
            pass
        except psutil.NoSuchProcess:
            pass
    return addresses


def find_gcs_addresses():
    """Finds any local GCS processes based on grepping ps."""
    return _find_address_from_flag("--gcs-address")


def find_bootstrap_address(temp_dir: Optional[str]):
    """Finds the latest Ray cluster address to connect to, if any. This is the
    GCS address connected to by the last successful `ray start`."""
    return ray._private.utils.read_ray_address(temp_dir)


def get_ray_address_from_environment(addr: str, temp_dir: Optional[str]):
    """Attempts to find the address of Ray cluster to use, in this order:

    1. Use RAY_ADDRESS if defined and nonempty.
    2. If no address is provided or the provided address is "auto", use the
    address in /tmp/ray/ray_current_cluster if available. This will error if
    the specified address is None and there is no address found. For "auto",
    we will fallback to connecting to any detected Ray cluster (legacy).
    3. Otherwise, use the provided address.

    Returns:
        A string to pass into `ray.init(address=...)`, e.g. ip:port, `auto`.
    """
    env_addr = os.environ.get(ray_constants.RAY_ADDRESS_ENVIRONMENT_VARIABLE)
    if env_addr is not None and env_addr != "":
        addr = env_addr

    if addr is not None and addr != "auto":
        return addr
    # We should try to automatically find an active local instance.
    gcs_addrs = find_gcs_addresses()
    bootstrap_addr = find_bootstrap_address(temp_dir)

    if len(gcs_addrs) > 1 and bootstrap_addr is not None:
        logger.warning(
            f"Found multiple active Ray instances: {gcs_addrs}. "
            f"Connecting to latest cluster at {bootstrap_addr}. "
            "You can override this by setting the `--address` flag "
            "or `RAY_ADDRESS` environment variable."
        )
    elif len(gcs_addrs) > 0 and addr == "auto":
        # Preserve legacy "auto" behavior of connecting to any cluster, even if not
        # started with ray start. However if addr is None, we will raise an error.
        bootstrap_addr = list(gcs_addrs).pop()

    if bootstrap_addr is None:
        if addr is None:
            # Caller should start a new instance.
            return None
        else:
            raise ConnectionError(
                "Could not find any running Ray instance. "
                "Please specify the one to connect to by setting `--address` flag "
                "or `RAY_ADDRESS` environment variable."
            )

    return bootstrap_addr


def wait_for_node(
    gcs_address: str,
    node_plasma_store_socket_name: str,
    timeout: int = _timeout,
):
    """Wait until this node has appeared in the client table.

    Args:
        gcs_address: The gcs address
        node_plasma_store_socket_name: The
            plasma_store_socket_name for the given node which we wait for.
        timeout: The amount of time in seconds to wait before raising an
            exception.

    Raises:
        TimeoutError: An exception is raised if the timeout expires before
            the node appears in the client table.
    """
    gcs_options = GcsClientOptions.from_gcs_address(gcs_address)
    global_state = ray._private.state.GlobalState()
    global_state._initialize_global_state(gcs_options)
    start_time = time.time()
    while time.time() - start_time < timeout:
        clients = global_state.node_table()
        object_store_socket_names = [
            client["ObjectStoreSocketName"] for client in clients
        ]
        if node_plasma_store_socket_name in object_store_socket_names:
            return
        else:
            time.sleep(0.1)
    raise TimeoutError(
        f"Timed out after {timeout} seconds while waiting for node to startup. "
        f"Did not find socket name {node_plasma_store_socket_name} in the list "
        "of object store socket names."
    )


def get_node_to_connect_for_driver(gcs_address, node_ip_address):
    # Get node table from global state accessor.
    global_state = ray._private.state.GlobalState()
    gcs_options = _get_gcs_client_options(gcs_address)
    global_state._initialize_global_state(gcs_options)
    return global_state.get_node_to_connect_for_driver(node_ip_address)


def get_webui_url_from_internal_kv():
    assert ray.experimental.internal_kv._internal_kv_initialized()
    webui_url = ray.experimental.internal_kv._internal_kv_get(
        "webui:url", namespace=ray_constants.KV_NAMESPACE_DASHBOARD
    )
    return ray._private.utils.decode(webui_url) if webui_url is not None else None


def get_storage_uri_from_internal_kv():
    assert ray.experimental.internal_kv._internal_kv_initialized()
    storage_uri = ray.experimental.internal_kv._internal_kv_get(
        "storage", namespace=ray_constants.KV_NAMESPACE_SESSION
    )
    return ray._private.utils.decode(storage_uri) if storage_uri is not None else None


def remaining_processes_alive():
    """See if the remaining processes are alive or not.

    Note that this ignores processes that have been explicitly killed,
    e.g., via a command like node.kill_raylet().

    Returns:
        True if the remaining processes started by ray.init() are alive and
            False otherwise.

    Raises:
        Exception: An exception is raised if the processes were not started by
            ray.init().
    """
    if ray._private.worker._global_node is None:
        raise RuntimeError(
            "This process is not in a position to determine "
            "whether all processes are alive or not."
        )
    return ray._private.worker._global_node.remaining_processes_alive()


def canonicalize_bootstrap_address(
    addr: str, temp_dir: Optional[str] = None
) -> Optional[str]:
    """Canonicalizes Ray cluster bootstrap address to host:port.
    Reads address from the environment if needed.

    This function should be used to process user supplied Ray cluster address,
    via ray.init() or `--address` flags, before using the address to connect.

    Returns:
        Ray cluster address string in <host:port> format or None if the caller
        should start a local Ray instance.
    """
    if addr is None or addr == "auto":
        addr = get_ray_address_from_environment(addr, temp_dir)
    if addr is None or addr == "local":
        return None
    try:
        bootstrap_address = resolve_ip_for_localhost(addr)
    except Exception:
        logger.exception(f"Failed to convert {addr} to host:port")
        raise
    return bootstrap_address


def canonicalize_bootstrap_address_or_die(
    addr: str, temp_dir: Optional[str] = None
) -> str:
    """Canonicalizes Ray cluster bootstrap address to host:port.

    This function should be used when the caller expects there to be an active
    and local Ray instance. If no address is provided or address="auto", this
    will autodetect the latest Ray instance created with `ray start`.

    For convenience, if no address can be autodetected, this function will also
    look for any running local GCS processes, based on pgrep output. This is to
    allow easier use of Ray CLIs when debugging a local Ray instance (whose GCS
    addresses are not recorded).

    Returns:
        Ray cluster address string in <host:port> format. Throws a
        ConnectionError if zero or multiple active Ray instances are
        autodetected.
    """
    bootstrap_addr = canonicalize_bootstrap_address(addr, temp_dir=temp_dir)
    if bootstrap_addr is not None:
        return bootstrap_addr

    running_gcs_addresses = find_gcs_addresses()
    if len(running_gcs_addresses) == 0:
        raise ConnectionError(
            "Could not find any running Ray instance. "
            "Please specify the one to connect to by setting the `--address` "
            "flag or `RAY_ADDRESS` environment variable."
        )
    if len(running_gcs_addresses) > 1:
        raise ConnectionError(
            f"Found multiple active Ray instances: {running_gcs_addresses}. "
            "Please specify the one to connect to by setting the `--address` "
            "flag or `RAY_ADDRESS` environment variable."
        )
    return running_gcs_addresses.pop()


def extract_ip_port(bootstrap_address: str):
    if ":" not in bootstrap_address:
        raise ValueError(
            f"Malformed address {bootstrap_address}. " f"Expected '<host>:<port>'."
        )
    ip, _, port = bootstrap_address.rpartition(":")
    try:
        port = int(port)
    except ValueError:
        raise ValueError(f"Malformed address port {port}. Must be an integer.")
    if port < 1024 or port > 65535:
        raise ValueError(
            f"Invalid address port {port}. Must be between 1024 "
            "and 65535 (inclusive)."
        )
    return ip, port


def resolve_ip_for_localhost(address: str):
    """Convert to a remotely reachable IP if the address is "localhost"
            or "127.0.0.1". Otherwise do nothing.

    Args:
        address: This can be either a string containing a hostname (or an IP
            address) and a port or it can be just an IP address.

    Returns:
        The same address but with the local host replaced by remotely
            reachable IP.
    """
    if not address:
        raise ValueError(f"Malformed address: {address}")
    address_parts = address.split(":")
    if address_parts[0] == "127.0.0.1" or address_parts[0] == "localhost":
        # Make sure localhost isn't resolved to the loopback ip
        ip_address = get_node_ip_address()
        return ":".join([ip_address] + address_parts[1:])
    else:
        return address


def node_ip_address_from_perspective(address: str):
    """IP address by which the local node can be reached *from* the `address`.

    Args:
        address: The IP address and port of any known live service on the
            network you care about.

    Returns:
        The IP address by which the local node can be reached from the address.
    """
    ip_address, port = address.split(":")
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    try:
        # This command will raise an exception if there is no internet
        # connection.
        s.connect((ip_address, int(port)))
        node_ip_address = s.getsockname()[0]
    except OSError as e:
        node_ip_address = "127.0.0.1"
        # [Errno 101] Network is unreachable
        if e.errno == errno.ENETUNREACH:
            try:
                # try get node ip address from host name
                host_name = socket.getfqdn(socket.gethostname())
                node_ip_address = socket.gethostbyname(host_name)
            except Exception:
                pass
    finally:
        s.close()

    return node_ip_address


def get_node_ip_address(address="8.8.8.8:53"):
    if ray._private.worker._global_node is not None:
        return ray._private.worker._global_node.node_ip_address
    if not ray_constants.ENABLE_RAY_CLUSTER:
        # Use loopback IP as the local IP address to prevent bothersome
        # firewall popups on OSX and Windows.
        # https://github.com/ray-project/ray/issues/18730.
        return "127.0.0.1"
    return node_ip_address_from_perspective(address)


def create_redis_client(redis_address, password=None):
    """Create a Redis client.

    Args:
        The IP address, port, and password of the Redis server.

    Returns:
        A Redis client.
    """
    import redis

    if not hasattr(create_redis_client, "instances"):
        create_redis_client.instances = {}

    num_retries = ray_constants.START_REDIS_WAIT_RETRIES
    delay = 0.001
    for i in range(num_retries):
        cli = create_redis_client.instances.get(redis_address)
        if cli is None:
            redis_ip_address, redis_port = extract_ip_port(
                canonicalize_bootstrap_address_or_die(redis_address)
            )
            cli = redis.StrictRedis(
                host=redis_ip_address, port=int(redis_port), password=password
            )
            create_redis_client.instances[redis_address] = cli
        try:
            cli.ping()
            return cli
        except Exception as e:
            create_redis_client.instances.pop(redis_address)
            if i >= num_retries - 1:
                raise RuntimeError(
                    f"Unable to connect to Redis at {redis_address}: {e}"
                )
            # Wait a little bit.
            time.sleep(delay)
            # Make sure the retry interval doesn't increase too large.
            delay = min(1, delay * 2)


def start_ray_process(
    command: List[str],
    process_type: str,
    fate_share: bool,
    env_updates: Optional[dict] = None,
    cwd: Optional[str] = None,
    use_valgrind: bool = False,
    use_gdb: bool = False,
    use_valgrind_profiler: bool = False,
    use_perftools_profiler: bool = False,
    use_tmux: bool = False,
    stdout_file: Optional[str] = None,
    stderr_file: Optional[str] = None,
    pipe_stdin: bool = False,
):
    """Start one of the Ray processes.

    TODO(rkn): We need to figure out how these commands interact. For example,
    it may only make sense to start a process in gdb if we also start it in
    tmux. Similarly, certain combinations probably don't make sense, like
    simultaneously running the process in valgrind and the profiler.

    Args:
        command: The command to use to start the Ray process.
        process_type: The type of the process that is being started
            (e.g., "raylet").
        fate_share: If true, the child will be killed if its parent (us) dies.
            True must only be passed after detection of this functionality.
        env_updates: A dictionary of additional environment variables to
            run the command with (in addition to the caller's environment
            variables).
        cwd: The directory to run the process in.
        use_valgrind: True if we should start the process in valgrind.
        use_gdb: True if we should start the process in gdb.
        use_valgrind_profiler: True if we should start the process in
            the valgrind profiler.
        use_perftools_profiler: True if we should profile the process
            using perftools.
        use_tmux: True if we should start the process in tmux.
        stdout_file: A file handle opened for writing to redirect stdout to. If
            no redirection should happen, then this should be None.
        stderr_file: A file handle opened for writing to redirect stderr to. If
            no redirection should happen, then this should be None.
        pipe_stdin: If true, subprocess.PIPE will be passed to the process as
            stdin.

    Returns:
        Information about the process that was started including a handle to
            the process that was started.
    """
    # Detect which flags are set through environment variables.
    valgrind_env_var = f"RAY_{process_type.upper()}_VALGRIND"
    if os.environ.get(valgrind_env_var) == "1":
        logger.info("Detected environment variable '%s'.", valgrind_env_var)
        use_valgrind = True
    valgrind_profiler_env_var = f"RAY_{process_type.upper()}_VALGRIND_PROFILER"
    if os.environ.get(valgrind_profiler_env_var) == "1":
        logger.info("Detected environment variable '%s'.", valgrind_profiler_env_var)
        use_valgrind_profiler = True
    perftools_profiler_env_var = f"RAY_{process_type.upper()}_PERFTOOLS_PROFILER"
    if os.environ.get(perftools_profiler_env_var) == "1":
        logger.info("Detected environment variable '%s'.", perftools_profiler_env_var)
        use_perftools_profiler = True
    tmux_env_var = f"RAY_{process_type.upper()}_TMUX"
    if os.environ.get(tmux_env_var) == "1":
        logger.info("Detected environment variable '%s'.", tmux_env_var)
        use_tmux = True
    gdb_env_var = f"RAY_{process_type.upper()}_GDB"
    if os.environ.get(gdb_env_var) == "1":
        logger.info("Detected environment variable '%s'.", gdb_env_var)
        use_gdb = True
    # Jemalloc memory profiling.
    jemalloc_lib_path = os.environ.get(RAY_JEMALLOC_LIB_PATH)
    jemalloc_conf = os.environ.get(RAY_JEMALLOC_CONF)
    jemalloc_comps = os.environ.get(RAY_JEMALLOC_PROFILE)
    jemalloc_comps = [] if not jemalloc_comps else jemalloc_comps.split(",")
    jemalloc_env_vars = propagate_jemalloc_env_var(
        jemalloc_path=jemalloc_lib_path,
        jemalloc_conf=jemalloc_conf,
        jemalloc_comps=jemalloc_comps,
        process_type=process_type,
    )
    use_jemalloc_mem_profiler = len(jemalloc_env_vars) > 0

    if (
        sum(
            [
                use_gdb,
                use_valgrind,
                use_valgrind_profiler,
                use_perftools_profiler,
                use_jemalloc_mem_profiler,
            ]
        )
        > 1
    ):
        raise ValueError(
            "At most one of the 'use_gdb', 'use_valgrind', "
            "'use_valgrind_profiler', 'use_perftools_profiler', "
            "and 'use_jemalloc_mem_profiler' flags can "
            "be used at a time."
        )
    if env_updates is None:
        env_updates = {}
    if not isinstance(env_updates, dict):
        raise ValueError("The 'env_updates' argument must be a dictionary.")

    modified_env = os.environ.copy()
    modified_env.update(env_updates)

    if use_gdb:
        if not use_tmux:
            raise ValueError(
                "If 'use_gdb' is true, then 'use_tmux' must be true as well."
            )

        # TODO(suquark): Any better temp file creation here?
        gdb_init_path = os.path.join(
            ray._private.utils.get_ray_temp_dir(),
            f"gdb_init_{process_type}_{time.time()}",
        )
        ray_process_path = command[0]
        ray_process_args = command[1:]
        run_args = " ".join(["'{}'".format(arg) for arg in ray_process_args])
        with open(gdb_init_path, "w") as gdb_init_file:
            gdb_init_file.write(f"run {run_args}")
        command = ["gdb", ray_process_path, "-x", gdb_init_path]

    if use_valgrind:
        command = [
            "valgrind",
            "--track-origins=yes",
            "--leak-check=full",
            "--show-leak-kinds=all",
            "--leak-check-heuristics=stdstring",
            "--error-exitcode=1",
        ] + command

    if use_valgrind_profiler:
        command = ["valgrind", "--tool=callgrind"] + command

    if use_perftools_profiler:
        modified_env["LD_PRELOAD"] = os.environ["PERFTOOLS_PATH"]
        modified_env["CPUPROFILE"] = os.environ["PERFTOOLS_LOGFILE"]

    if use_jemalloc_mem_profiler:
        logger.info(
            f"Jemalloc profiling will be used for {process_type}. "
            f"env vars: {jemalloc_env_vars}"
        )
        modified_env.update(jemalloc_env_vars)

    if use_tmux:
        # The command has to be created exactly as below to ensure that it
        # works on all versions of tmux. (Tested with tmux 1.8-5, travis'
        # version, and tmux 2.1)
        command = ["tmux", "new-session", "-d", f"{' '.join(command)}"]

    if fate_share:
        assert ray._private.utils.detect_fate_sharing_support(), (
            "kernel-level fate-sharing must only be specified if "
            "detect_fate_sharing_support() has returned True"
        )

    def preexec_fn():
        import signal

        signal.pthread_sigmask(signal.SIG_BLOCK, {signal.SIGINT})
        if fate_share and sys.platform.startswith("linux"):
            ray._private.utils.set_kill_on_parent_death_linux()

    win32_fate_sharing = fate_share and sys.platform == "win32"
    # With Windows fate-sharing, we need special care:
    # The process must be added to the job before it is allowed to execute.
    # Otherwise, there's a race condition: the process might spawn children
    # before the process itself is assigned to the job.
    # After that point, its children will not be added to the job anymore.
    CREATE_SUSPENDED = 0x00000004  # from Windows headers
    if sys.platform == "win32":
        # CreateProcess, which underlies Popen, is limited to
        # 32,767 characters, including the Unicode terminating null
        # character
        total_chrs = sum([len(x) for x in command])
        if total_chrs > 31766:
            raise ValueError(
                f"command is limited to a total of 31767 characters, "
                f"got {total_chrs}"
            )

    process = ConsolePopen(
        command,
        env=modified_env,
        cwd=cwd,
        stdout=stdout_file,
        stderr=stderr_file,
        stdin=subprocess.PIPE if pipe_stdin else None,
        preexec_fn=preexec_fn if sys.platform != "win32" else None,
        creationflags=CREATE_SUSPENDED if win32_fate_sharing else 0,
    )

    if win32_fate_sharing:
        try:
            ray._private.utils.set_kill_child_on_death_win32(process)
            psutil.Process(process.pid).resume()
        except (psutil.Error, OSError):
            process.kill()
            raise

    def _get_stream_name(stream):
        if stream is not None:
            try:
                return stream.name
            except AttributeError:
                return str(stream)
        return None

    return ProcessInfo(
        process=process,
        stdout_file=_get_stream_name(stdout_file),
        stderr_file=_get_stream_name(stderr_file),
        use_valgrind=use_valgrind,
        use_gdb=use_gdb,
        use_valgrind_profiler=use_valgrind_profiler,
        use_perftools_profiler=use_perftools_profiler,
        use_tmux=use_tmux,
    )


def start_reaper(fate_share=None):
    """Start the reaper process.

    This is a lightweight process that simply
    waits for its parent process to die and then terminates its own
    process group. This allows us to ensure that ray processes are always
    terminated properly so long as that process itself isn't SIGKILLed.

    Returns:
        ProcessInfo for the process that was started.
    """
    # Make ourselves a process group leader so that the reaper can clean
    # up other ray processes without killing the process group of the
    # process that started us.
    try:
        if sys.platform != "win32":
            os.setpgrp()
    except OSError as e:
        errcode = e.errno
        if errcode == errno.EPERM and os.getpgrp() == os.getpid():
            # Nothing to do; we're already a session leader.
            pass
        else:
            logger.warning(
                f"setpgrp failed, processes may not be cleaned up properly: {e}."
            )
            # Don't start the reaper in this case as it could result in killing
            # other user processes.
            return None

    reaper_filepath = os.path.join(RAY_PATH, RAY_PRIVATE_DIR, "ray_process_reaper.py")
    command = [sys.executable, "-u", reaper_filepath]
    process_info = start_ray_process(
        command,
        ray_constants.PROCESS_TYPE_REAPER,
        pipe_stdin=True,
        fate_share=fate_share,
    )
    return process_info


def start_log_monitor(
    logs_dir: str,
    gcs_address: str,
    fate_share: Optional[bool] = None,
    max_bytes: int = 0,
    backup_count: int = 0,
    redirect_logging: bool = True,
    stdout_file: Optional[IO[AnyStr]] = subprocess.DEVNULL,
    stderr_file: Optional[IO[AnyStr]] = subprocess.DEVNULL,
):
    """Start a log monitor process.

    Args:
        logs_dir: The directory of logging files.
        gcs_address: GCS address for pubsub.
        fate_share: Whether to share fate between log_monitor
            and this process.
        max_bytes: Log rotation parameter. Corresponding to
            RotatingFileHandler's maxBytes.
        backup_count: Log rotation parameter. Corresponding to
            RotatingFileHandler's backupCount.
        redirect_logging: Whether we should redirect logging to
            the provided log directory.
        stdout_file: A file handle opened for writing to redirect stdout to. If
            no redirection should happen, then this should be None.
        stderr_file: A file handle opened for writing to redirect stderr to. If
            no redirection should happen, then this should be None.

    Returns:
        ProcessInfo for the process that was started.
    """
    log_monitor_filepath = os.path.join(RAY_PATH, RAY_PRIVATE_DIR, "log_monitor.py")

    command = [
        sys.executable,
        "-u",
        log_monitor_filepath,
        f"--logs-dir={logs_dir}",
        f"--gcs-address={gcs_address}",
        f"--logging-rotate-bytes={max_bytes}",
        f"--logging-rotate-backup-count={backup_count}",
    ]

    if not redirect_logging:
        # If not redirecting logging to files, unset log filename.
        # This will cause log records to go to stderr.
        command.append("--logging-filename=")
        # Use stderr log format with the component name as a message prefix.
        logging_format = ray_constants.LOGGER_FORMAT_STDERR.format(
            component=ray_constants.PROCESS_TYPE_LOG_MONITOR
        )
        command.append(f"--logging-format={logging_format}")
        # Inherit stdout/stderr streams.
        stdout_file = None
        stderr_file = None
    process_info = start_ray_process(
        command,
        ray_constants.PROCESS_TYPE_LOG_MONITOR,
        stdout_file=stdout_file,
        stderr_file=stderr_file,
        fate_share=fate_share,
    )
    return process_info


def start_api_server(
    include_dashboard: bool,
    raise_on_failure: bool,
    host: str,
    gcs_address: str,
    node_ip_address: str,
    temp_dir: str,
    logdir: str,
    session_dir: str,
    port: Optional[int] = None,
    dashboard_grpc_port: Optional[int] = None,
    fate_share: Optional[bool] = None,
    max_bytes: int = 0,
    backup_count: int = 0,
    redirect_logging: bool = True,
    stdout_file: Optional[IO[AnyStr]] = subprocess.DEVNULL,
    stderr_file: Optional[IO[AnyStr]] = subprocess.DEVNULL,
):
    """Start a API server process.

    Args:
        include_dashboard: If true, this will load all dashboard-related modules
            when starting the API server. Otherwise, it will only
            start the modules that are not relevant to the dashboard.
        raise_on_failure: If true, this will raise an exception
            if we fail to start the API server. Otherwise it will print
            a warning if we fail to start the API server.
        host: The host to bind the dashboard web server to.
        gcs_address: The gcs address the dashboard should connect to
        node_ip_address: The IP address where this is running.
        temp_dir: The temporary directory used for log files and
            information for this Ray session.
        session_dir: The session directory under temp_dir.
            It is used as a identifier of individual cluster.
        logdir: The log directory used to generate dashboard log.
        port: The port to bind the dashboard web server to.
            Defaults to 8265.
        dashboard_grpc_port: The port which the dashboard listens for
            gRPC on. Defaults to a random, available port.
        max_bytes: Log rotation parameter. Corresponding to
            RotatingFileHandler's maxBytes.
        backup_count: Log rotation parameter. Corresponding to
            RotatingFileHandler's backupCount.
        redirect_logging: Whether we should redirect logging to
            the provided log directory.
        stdout_file: A file handle opened for writing to redirect stdout to. If
            no redirection should happen, then this should be None.
        stderr_file: A file handle opened for writing to redirect stderr to. If
            no redirection should happen, then this should be None.

    Returns:
        A tuple of :
            - Dashboard URL if dashboard enabled and started.
            - ProcessInfo for the process that was started.
    """
    try:
        # Make sure port is available.
        if port is None:
            port_retries = 50
            port = ray_constants.DEFAULT_DASHBOARD_PORT
        else:
            port_retries = 0
            port_test_socket = socket.socket()
            port_test_socket.setsockopt(
                socket.SOL_SOCKET,
                socket.SO_REUSEADDR,
                1,
            )
            try:
                port_test_socket.bind((host, port))
                port_test_socket.close()
            except socket.error as e:
                # 10013 on windows is a bit more broad than just
                # "address in use": it can also indicate "permission denied".
                # TODO: improve the error message?
                if e.errno in {48, 98, 10013}:  # address already in use.
                    raise ValueError(
                        f"Failed to bind to {host}:{port} because it's "
                        "already occupied. You can use `ray start "
                        "--dashboard-port ...` or `ray.init(dashboard_port=..."
                        ")` to select a different port."
                    )
                else:
                    raise e
        # Make sure the process can start.
        minimal = not ray._private.utils.check_dashboard_dependencies_installed()
        # Start the dashboard process.
        dashboard_dir = "dashboard"
        dashboard_filepath = os.path.join(RAY_PATH, dashboard_dir, "dashboard.py")

        command = [
            *_build_python_executable_command_memory_profileable(
                ray_constants.PROCESS_TYPE_DASHBOARD,
                session_dir,
                unbuffered=False,
            ),
            dashboard_filepath,
            f"--host={host}",
            f"--port={port}",
            f"--port-retries={port_retries}",
            f"--temp-dir={temp_dir}",
            f"--log-dir={logdir}",
            f"--session-dir={session_dir}",
            f"--logging-rotate-bytes={max_bytes}",
            f"--logging-rotate-backup-count={backup_count}",
            f"--gcs-address={gcs_address}",
            f"--node-ip-address={node_ip_address}",
        ]

        if not redirect_logging:
            # If not redirecting logging to files, unset log filename.
            # This will cause log records to go to stderr.
            command.append("--logging-filename=")
            # Use stderr log format with the component name as a message prefix.
            logging_format = ray_constants.LOGGER_FORMAT_STDERR.format(
                component=ray_constants.PROCESS_TYPE_DASHBOARD
            )
            command.append(f"--logging-format={logging_format}")
            # Inherit stdout/stderr streams so that
            # logs are redirected to stderr.
            stdout_file = None
            stderr_file = None
        if minimal:
            command.append("--minimal")

        if not include_dashboard:
            # If dashboard is not included, load modules
            # that are irrelevant to the dashboard.
            # TODO(sang): Modules like job or state APIs should be
            # loaded although dashboard is disabled. Fix it.
            command.append("--modules-to-load=UsageStatsHead")
            command.append("--disable-frontend")

        if dashboard_grpc_port is not None:
            command.append(f"--grpc-port={dashboard_grpc_port}")

        process_info = start_ray_process(
            command,
            ray_constants.PROCESS_TYPE_DASHBOARD,
            stdout_file=stdout_file,
            stderr_file=stderr_file,
            fate_share=fate_share,
        )

        # Retrieve the dashboard url
        gcs_client = GcsClient(address=gcs_address)
        ray.experimental.internal_kv._initialize_internal_kv(gcs_client)
        dashboard_url = None
        dashboard_returncode = None
        for _ in range(200):
            dashboard_url = ray.experimental.internal_kv._internal_kv_get(
                ray_constants.DASHBOARD_ADDRESS,
                namespace=ray_constants.KV_NAMESPACE_DASHBOARD,
            )
            if dashboard_url is not None:
                dashboard_url = dashboard_url.decode("utf-8")
                break
            dashboard_returncode = process_info.process.poll()
            if dashboard_returncode is not None:
                break
            # This is often on the critical path of ray.init() and ray start,
            # so we need to poll often.
            time.sleep(0.1)

        # Dashboard couldn't be started.
        if dashboard_url is None:
            returncode_str = (
                f", return code {dashboard_returncode}"
                if dashboard_returncode is not None
                else ""
            )
            logger.error(f"Failed to start the dashboard {returncode_str}")

            def read_log(filename, lines_to_read):
                """Read a log file and return the last 20 lines."""
                dashboard_log = os.path.join(logdir, filename)
                # Read last n lines of dashboard log. The log file may be large.
                lines_to_read = 20
                lines = []
                with open(dashboard_log, "rb") as f:
                    with mmap.mmap(f.fileno(), 0, access=mmap.ACCESS_READ) as mm:
                        end = mm.size()
                        for _ in range(lines_to_read):
                            sep = mm.rfind(b"\n", 0, end - 1)
                            if sep == -1:
                                break
                            lines.append(mm[sep + 1 : end].decode("utf-8"))
                            end = sep
                lines.append(
                    f"The last {lines_to_read} lines of {dashboard_log} "
                    "(it contains the error message from the dashboard): "
                )
                return lines

            if logdir:
                lines_to_read = 20
                logger.error(
                    "Error should be written to 'dashboard.log' or "
                    "'dashboard.err'. We are printing the last "
                    f"{lines_to_read} lines for you. See "
                    "'https://docs.ray.io/en/master/ray-observability/ray-logging.html#logging-directory-structure' "  # noqa
                    "to find where the log file is."
                )
                try:
                    lines = read_log("dashboard.log", lines_to_read=lines_to_read)
                except Exception as e:
                    logger.error(
                        f"Couldn't read dashboard.log file. Error: {e}. "
                        "It means the dashboard is broken even before it "
                        "initializes the logger (mostly dependency issues). "
                        "Reading the dashboard.err file which contains stdout/stderr."
                    )
                    # If we cannot read the .log file, we fallback to .err file.
                    # This is the case where dashboard couldn't be started at all
                    # and couldn't even initialize the logger to write logs to .log
                    # file.
                    try:
                        lines = read_log("dashboard.err", lines_to_read=lines_to_read)
                    except Exception as e:
                        raise Exception(
                            f"Failed to read dashboard.err file: {e}. "
                            "It is unexpected. Please report an issue to "
                            "Ray github. "
                            "https://github.com/ray-project/ray/issues"
                        )
                last_log_str = "\n" + "\n".join(reversed(lines[-lines_to_read:]))
                raise Exception(last_log_str)
            else:
                # Is it reachable?
                raise Exception("Failed to start a dashboard.")

        if minimal or not include_dashboard:
            # If it is the minimal installation, the web url (dashboard url)
            # shouldn't be configured because it doesn't start a server.
            dashboard_url = ""
        return dashboard_url, process_info
    except Exception as e:
        if raise_on_failure:
            raise e from e
        else:
            logger.error(e)
            return None, None


def start_gcs_server(
    redis_address: str,
    log_dir: str,
    session_name: str,
    stdout_file: Optional[str] = None,
    stderr_file: Optional[str] = None,
    redis_password: Optional[str] = None,
    config: Optional[dict] = None,
    fate_share: Optional[bool] = None,
    gcs_server_port: Optional[int] = None,
    metrics_agent_port: Optional[int] = None,
    node_ip_address: Optional[str] = None,
):
    """Start a gcs server.

    Args:
        redis_address: The address that the Redis server is listening on.
        log_dir: The path of the dir where log files are created.
        session_name: The session name (cluster id) of this cluster.
        stdout_file: A file handle opened for writing to redirect stdout to. If
            no redirection should happen, then this should be None.
        stderr_file: A file handle opened for writing to redirect stderr to. If
            no redirection should happen, then this should be None.
        redis_password: The password of the redis server.
        config: Optional configuration that will
            override defaults in RayConfig.
        gcs_server_port: Port number of the gcs server.
        metrics_agent_port: The port where metrics agent is bound to.
        node_ip_address: IP Address of a node where gcs server starts.

    Returns:
        ProcessInfo for the process that was started.
    """
    assert gcs_server_port > 0

    command = [
        GCS_SERVER_EXECUTABLE,
        f"--log_dir={log_dir}",
        f"--config_list={serialize_config(config)}",
        f"--gcs_server_port={gcs_server_port}",
        f"--metrics-agent-port={metrics_agent_port}",
        f"--node-ip-address={node_ip_address}",
        f"--session-name={session_name}",
    ]
    if redis_address:
        parts = redis_address.split("://", 1)
        enable_redis_ssl = "false"
        if len(parts) == 1:
            redis_ip_address, redis_port = parts[0].rsplit(":", 1)
        else:
            if len(parts) != 2 or parts[0] not in ("redis", "rediss"):
                raise ValueError(f"Invalid redis address {redis_address}")
            redis_ip_address, redis_port = parts[1].rsplit(":", 1)
            if parts[0] == "rediss":
                enable_redis_ssl = "true"

        command += [
            f"--redis_address={redis_ip_address}",
            f"--redis_port={redis_port}",
            f"--redis_enable_ssl={enable_redis_ssl}",
        ]
    if redis_password:
        command += [f"--redis_password={redis_password}"]
    process_info = start_ray_process(
        command,
        ray_constants.PROCESS_TYPE_GCS_SERVER,
        stdout_file=stdout_file,
        stderr_file=stderr_file,
        fate_share=fate_share,
    )
    return process_info


def start_raylet(
    redis_address: str,
    gcs_address: str,
    node_ip_address: str,
    node_manager_port: int,
    raylet_name: str,
    plasma_store_name: str,
    worker_path: str,
    setup_worker_path: str,
    storage: str,
    temp_dir: str,
    session_dir: str,
    resource_dir: str,
    log_dir: str,
    resource_spec,
    plasma_directory: str,
    object_store_memory: int,
    session_name: str,
    is_head_node: bool,
    min_worker_port: Optional[int] = None,
    max_worker_port: Optional[int] = None,
    worker_port_list: Optional[List[int]] = None,
    object_manager_port: Optional[int] = None,
    redis_password: Optional[str] = None,
    metrics_agent_port: Optional[int] = None,
    metrics_export_port: Optional[int] = None,
    dashboard_agent_listen_port: Optional[int] = None,
    use_valgrind: bool = False,
    use_profiler: bool = False,
    stdout_file: Optional[str] = None,
    stderr_file: Optional[str] = None,
    config: Optional[dict] = None,
    huge_pages: bool = False,
    fate_share: Optional[bool] = None,
    socket_to_use: Optional[int] = None,
    max_bytes: int = 0,
    backup_count: int = 0,
    ray_debugger_external: bool = False,
    env_updates: Optional[dict] = None,
    node_name: Optional[str] = None,
    webui: Optional[str] = None,
    labels: Optional[dict] = None,
):
    """Start a raylet, which is a combined local scheduler and object manager.

    Args:
        redis_address: The address of the primary Redis server.
        gcs_address: The address of GCS server.
        node_ip_address: The IP address of this node.
        node_manager_port: The port to use for the node manager. If it's
            0, a random port will be used.
        raylet_name: The name of the raylet socket to create.
        plasma_store_name: The name of the plasma store socket to connect
             to.
        worker_path: The path of the Python file that new worker
            processes will execute.
        setup_worker_path: The path of the Python file that will set up
            the environment for the worker process.
        storage: The persistent storage URI.
        temp_dir: The path of the temporary directory Ray will use.
        session_dir: The path of this session.
        resource_dir: The path of resource of this session .
        log_dir: The path of the dir where log files are created.
        resource_spec: Resources for this raylet.
        session_name: The session name (cluster id) of this cluster.
        object_manager_port: The port to use for the object manager. If this is
            None, then the object manager will choose its own port.
        min_worker_port: The lowest port number that workers will bind
            on. If not set, random ports will be chosen.
        max_worker_port: The highest port number that workers will bind
            on. If set, min_worker_port must also be set.
        redis_password: The password to use when connecting to Redis.
        metrics_agent_port: The port where metrics agent is bound to.
        metrics_export_port: The port at which metrics are exposed to.
        dashboard_agent_listen_port: The port at which the dashboard agent
            listens to for http.
        use_valgrind: True if the raylet should be started inside
            of valgrind. If this is True, use_profiler must be False.
        use_profiler: True if the raylet should be started inside
            a profiler. If this is True, use_valgrind must be False.
        stdout_file: A file handle opened for writing to redirect stdout to. If
            no redirection should happen, then this should be None.
        stderr_file: A file handle opened for writing to redirect stderr to. If
            no redirection should happen, then this should be None.
        tracing_startup_hook: Tracing startup hook.
        config: Optional Raylet configuration that will
            override defaults in RayConfig.
        max_bytes: Log rotation parameter. Corresponding to
            RotatingFileHandler's maxBytes.
        backup_count: Log rotation parameter. Corresponding to
            RotatingFileHandler's backupCount.
        ray_debugger_external: True if the Ray debugger should be made
            available externally to this node.
        env_updates: Environment variable overrides.
        labels: The key-value labels of the node.
    Returns:
        ProcessInfo for the process that was started.
    """
    assert node_manager_port is not None and type(node_manager_port) == int

    if use_valgrind and use_profiler:
        raise ValueError("Cannot use valgrind and profiler at the same time.")

    assert resource_spec.resolved()
    static_resources = resource_spec.to_resource_dict()

    # Limit the number of workers that can be started in parallel by the
    # raylet. However, make sure it is at least 1.
    num_cpus_static = static_resources.get("CPU", 0)
    maximum_startup_concurrency = max(
        1, min(multiprocessing.cpu_count(), num_cpus_static)
    )

    # Format the resource argument in a form like 'CPU,1.0,GPU,0,Custom,3'.
    resource_argument = ",".join(
        ["{},{}".format(*kv) for kv in static_resources.items()]
    )

    has_java_command = False
    if shutil.which("java") is not None:
        has_java_command = True

    ray_java_installed = False
    try:
        jars_dir = get_ray_jars_dir()
        if os.path.exists(jars_dir):
            ray_java_installed = True
    except Exception:
        pass

    include_java = has_java_command and ray_java_installed
    if include_java is True:
        java_worker_command = build_java_worker_command(
            gcs_address,
            plasma_store_name,
            raylet_name,
            redis_password,
            session_dir,
            node_ip_address,
            setup_worker_path,
        )
    else:
        java_worker_command = []

    if os.path.exists(DEFAULT_WORKER_EXECUTABLE):
        cpp_worker_command = build_cpp_worker_command(
            gcs_address,
            plasma_store_name,
            raylet_name,
            redis_password,
            session_dir,
            log_dir,
            node_ip_address,
            setup_worker_path,
        )
    else:
        cpp_worker_command = []

    # Create the command that the Raylet will use to start workers.
    # TODO(architkulkarni): Pipe in setup worker args separately instead of
    # inserting them into start_worker_command and later erasing them if
    # needed.
    start_worker_command = (
        [
            sys.executable,
            setup_worker_path,
        ]
        + _site_flags()  # Inherit "-S" and "-s" flags from current Python interpreter.
        + [
            worker_path,
            f"--node-ip-address={node_ip_address}",
            "--node-manager-port=RAY_NODE_MANAGER_PORT_PLACEHOLDER",
            f"--object-store-name={plasma_store_name}",
            f"--raylet-name={raylet_name}",
            f"--redis-address={redis_address}",
            f"--temp-dir={temp_dir}",
            f"--metrics-agent-port={metrics_agent_port}",
            f"--logging-rotate-bytes={max_bytes}",
            f"--logging-rotate-backup-count={backup_count}",
            f"--gcs-address={gcs_address}",
            f"--session-name={session_name}",
            f"--temp-dir={temp_dir}",
            f"--webui={webui}",
        ]
    )

    if storage is not None:
        start_worker_command.append(f"--storage={storage}")

    start_worker_command.append("RAY_WORKER_DYNAMIC_OPTION_PLACEHOLDER")

    if redis_password:
        start_worker_command += [f"--redis-password={redis_password}"]

    # If the object manager port is None, then use 0 to cause the object
    # manager to choose its own port.
    if object_manager_port is None:
        object_manager_port = 0

    if min_worker_port is None:
        min_worker_port = 0

    if max_worker_port is None:
        max_worker_port = 0

    labels_json_str = ""
    if labels:
        labels_json_str = json.dumps(labels)

    agent_command = [
        *_build_python_executable_command_memory_profileable(
            ray_constants.PROCESS_TYPE_DASHBOARD_AGENT, session_dir
        ),
        os.path.join(RAY_PATH, "dashboard", "agent.py"),
        f"--node-ip-address={node_ip_address}",
        f"--metrics-export-port={metrics_export_port}",
        f"--dashboard-agent-port={metrics_agent_port}",
        f"--listen-port={dashboard_agent_listen_port}",
        "--node-manager-port=RAY_NODE_MANAGER_PORT_PLACEHOLDER",
        f"--object-store-name={plasma_store_name}",
        f"--raylet-name={raylet_name}",
        f"--temp-dir={temp_dir}",
        f"--session-dir={session_dir}",
        f"--runtime-env-dir={resource_dir}",
        f"--log-dir={log_dir}",
        f"--logging-rotate-bytes={max_bytes}",
        f"--logging-rotate-backup-count={backup_count}",
        f"--session-name={session_name}",
        f"--gcs-address={gcs_address}",
    ]
    if stdout_file is None and stderr_file is None:
        # If not redirecting logging to files, unset log filename.
        # This will cause log records to go to stderr.
        agent_command.append("--logging-filename=")
        # Use stderr log format with the component name as a message prefix.
        logging_format = ray_constants.LOGGER_FORMAT_STDERR.format(
            component=ray_constants.PROCESS_TYPE_DASHBOARD_AGENT
        )
        agent_command.append(f"--logging-format={logging_format}")

    if not ray._private.utils.check_dashboard_dependencies_installed():
        # If dependencies are not installed, it is the minimally packaged
        # ray. We should restrict the features within dashboard agent
        # that requires additional dependencies to be downloaded.
        agent_command.append("--minimal")

    command = [
        RAYLET_EXECUTABLE,
        f"--raylet_socket_name={raylet_name}",
        f"--store_socket_name={plasma_store_name}",
        f"--object_manager_port={object_manager_port}",
        f"--min_worker_port={min_worker_port}",
        f"--max_worker_port={max_worker_port}",
        f"--node_manager_port={node_manager_port}",
        f"--node_ip_address={node_ip_address}",
        f"--maximum_startup_concurrency={maximum_startup_concurrency}",
        f"--static_resource_list={resource_argument}",
        f"--python_worker_command={subprocess.list2cmdline(start_worker_command)}",  # noqa
        f"--java_worker_command={subprocess.list2cmdline(java_worker_command)}",  # noqa
        f"--cpp_worker_command={subprocess.list2cmdline(cpp_worker_command)}",  # noqa
        f"--native_library_path={DEFAULT_NATIVE_LIBRARY_PATH}",
        f"--temp_dir={temp_dir}",
        f"--session_dir={session_dir}",
        f"--log_dir={log_dir}",
        f"--resource_dir={resource_dir}",
        f"--metrics-agent-port={metrics_agent_port}",
        f"--metrics_export_port={metrics_export_port}",
        f"--object_store_memory={object_store_memory}",
        f"--plasma_directory={plasma_directory}",
        f"--ray-debugger-external={1 if ray_debugger_external else 0}",
        f"--gcs-address={gcs_address}",
        f"--session-name={session_name}",
        f"--labels={labels_json_str}",
    ]

    if is_head_node:
        command.append("--head")

    if worker_port_list is not None:
        command.append(f"--worker_port_list={worker_port_list}")
    command.append(
        "--num_prestart_python_workers={}".format(int(resource_spec.num_cpus))
    )
    command.append("--agent_command={}".format(subprocess.list2cmdline(agent_command)))
    if huge_pages:
        command.append("--huge_pages")
    if socket_to_use:
        socket_to_use.close()
    if node_name is not None:
        command.append(
            f"--node-name={node_name}",
        )
    process_info = start_ray_process(
        command,
        ray_constants.PROCESS_TYPE_RAYLET,
        use_valgrind=use_valgrind,
        use_gdb=False,
        use_valgrind_profiler=use_profiler,
        use_perftools_profiler=("RAYLET_PERFTOOLS_PATH" in os.environ),
        stdout_file=stdout_file,
        stderr_file=stderr_file,
        fate_share=fate_share,
        env_updates=env_updates,
    )

    return process_info


def get_ray_jars_dir():
    """Return a directory where all ray-related jars and
    their dependencies locate."""
    current_dir = RAY_PATH
    jars_dir = os.path.abspath(os.path.join(current_dir, "jars"))
    if not os.path.exists(jars_dir):
        raise RuntimeError(
            "Ray jars is not packaged into ray. "
            "Please build ray with java enabled "
            "(set env var RAY_INSTALL_JAVA=1)"
        )
    return os.path.abspath(os.path.join(current_dir, "jars"))


def build_java_worker_command(
    bootstrap_address: str,
    plasma_store_name: str,
    raylet_name: str,
    redis_password: str,
    session_dir: str,
    node_ip_address: str,
    setup_worker_path: str,
):
    """This method assembles the command used to start a Java worker.

    Args:
        bootstrap_address: Bootstrap address of ray cluster.
        plasma_store_name: The name of the plasma store socket to connect
           to.
        raylet_name: The name of the raylet socket to create.
        redis_password: The password of connect to redis.
        session_dir: The path of this session.
        node_ip_address: The ip address for this node.
        setup_worker_path: The path of the Python file that will set up
            the environment for the worker process.
    Returns:
        The command string for starting Java worker.
    """
    pairs = []
    if bootstrap_address is not None:
        pairs.append(("ray.address", bootstrap_address))
    pairs.append(("ray.raylet.node-manager-port", "RAY_NODE_MANAGER_PORT_PLACEHOLDER"))

    if plasma_store_name is not None:
        pairs.append(("ray.object-store.socket-name", plasma_store_name))

    if raylet_name is not None:
        pairs.append(("ray.raylet.socket-name", raylet_name))

    if redis_password is not None:
        pairs.append(("ray.redis.password", redis_password))

    if node_ip_address is not None:
        pairs.append(("ray.node-ip", node_ip_address))

    pairs.append(("ray.home", RAY_HOME))
    pairs.append(("ray.logging.dir", os.path.join(session_dir, "logs")))
    pairs.append(("ray.session-dir", session_dir))
    command = (
        [sys.executable]
        + [setup_worker_path]
        + ["-D{}={}".format(*pair) for pair in pairs]
    )

    command += ["RAY_WORKER_DYNAMIC_OPTION_PLACEHOLDER"]
    command += ["io.ray.runtime.runner.worker.DefaultWorker"]

    return command


def build_cpp_worker_command(
    bootstrap_address: str,
    plasma_store_name: str,
    raylet_name: str,
    redis_password: str,
    session_dir: str,
    log_dir: str,
    node_ip_address: str,
    setup_worker_path: str,
):
    """This method assembles the command used to start a CPP worker.

    Args:
        bootstrap_address: The bootstrap address of the cluster.
        plasma_store_name: The name of the plasma store socket to connect
           to.
        raylet_name: The name of the raylet socket to create.
        redis_password: The password of connect to redis.
        session_dir: The path of this session.
        log_dir: The path of logs.
        node_ip_address: The ip address for this node.
        setup_worker_path: The path of the Python file that will set up
            the environment for the worker process.
    Returns:
        The command string for starting CPP worker.
    """

    command = [
        sys.executable,
        setup_worker_path,
        DEFAULT_WORKER_EXECUTABLE,
        f"--ray_plasma_store_socket_name={plasma_store_name}",
        f"--ray_raylet_socket_name={raylet_name}",
        "--ray_node_manager_port=RAY_NODE_MANAGER_PORT_PLACEHOLDER",
        f"--ray_address={bootstrap_address}",
        f"--ray_redis_password={redis_password}",
        f"--ray_session_dir={session_dir}",
        f"--ray_logs_dir={log_dir}",
        f"--ray_node_ip_address={node_ip_address}",
        "RAY_WORKER_DYNAMIC_OPTION_PLACEHOLDER",
    ]

    return command


def determine_plasma_store_config(
    object_store_memory: int,
    plasma_directory: Optional[str] = None,
    huge_pages: bool = False,
):
    """Figure out how to configure the plasma object store.

    This will determine which directory to use for the plasma store. On Linux,
    we will try to use /dev/shm unless the shared memory file system is too
    small, in which case we will fall back to /tmp. If any of the object store
    memory or plasma directory parameters are specified by the user, then those
    values will be preserved.

    Args:
        object_store_memory: The object store memory to use.
        plasma_directory: The user-specified plasma directory parameter.
        huge_pages: The user-specified huge pages parameter.

    Returns:
        The plasma directory to use. If it is specified by the user, then that
            value will be preserved.
    """
    if not isinstance(object_store_memory, int):
        object_store_memory = int(object_store_memory)

    if huge_pages and not (sys.platform == "linux" or sys.platform == "linux2"):
        raise ValueError("The huge_pages argument is only supported on Linux.")

    system_memory = ray._private.utils.get_system_memory()

    # Determine which directory to use. By default, use /tmp on MacOS and
    # /dev/shm on Linux, unless the shared-memory file system is too small,
    # in which case we default to /tmp on Linux.
    if plasma_directory is None:
        if sys.platform == "linux" or sys.platform == "linux2":
            shm_avail = ray._private.utils.get_shared_memory_bytes()
            # Compare the requested memory size to the memory available in
            # /dev/shm.
            if shm_avail > object_store_memory:
                plasma_directory = "/dev/shm"
            elif (
                not os.environ.get("RAY_OBJECT_STORE_ALLOW_SLOW_STORAGE")
                and object_store_memory > ray_constants.REQUIRE_SHM_SIZE_THRESHOLD
            ):
                raise ValueError(
                    "The configured object store size ({} GB) exceeds "
                    "/dev/shm size ({} GB). This will harm performance. "
                    "Consider deleting files in /dev/shm or increasing its "
                    "size with "
                    "--shm-size in Docker. To ignore this warning, "
                    "set RAY_OBJECT_STORE_ALLOW_SLOW_STORAGE=1.".format(
                        object_store_memory / 1e9, shm_avail / 1e9
                    )
                )
            else:
                plasma_directory = ray._private.utils.get_user_temp_dir()
                logger.warning(
                    "WARNING: The object store is using {} instead of "
                    "/dev/shm because /dev/shm has only {} bytes available. "
                    "This will harm performance! You may be able to free up "
                    "space by deleting files in /dev/shm. If you are inside a "
                    "Docker container, you can increase /dev/shm size by "
                    "passing '--shm-size={:.2f}gb' to 'docker run' (or add it "
                    "to the run_options list in a Ray cluster config). Make "
                    "sure to set this to more than 30% of available RAM.".format(
                        ray._private.utils.get_user_temp_dir(),
                        shm_avail,
                        object_store_memory * (1.1) / (2**30),
                    )
                )
        else:
            plasma_directory = ray._private.utils.get_user_temp_dir()

        # Do some sanity checks.
        if object_store_memory > system_memory:
            raise ValueError(
                "The requested object store memory size is greater "
                "than the total available memory."
            )
    else:
        plasma_directory = os.path.abspath(plasma_directory)
        logger.info("object_store_memory is not verified when plasma_directory is set.")

    if not os.path.isdir(plasma_directory):
        raise ValueError(
            f"The file {plasma_directory} does not exist or is not a directory."
        )

    if huge_pages and plasma_directory is None:
        raise ValueError(
            "If huge_pages is True, then the "
            "plasma_directory argument must be provided."
        )

    if object_store_memory < ray_constants.OBJECT_STORE_MINIMUM_MEMORY_BYTES:
        raise ValueError(
            "Attempting to cap object store memory usage at {} "
            "bytes, but the minimum allowed is {} bytes.".format(
                object_store_memory, ray_constants.OBJECT_STORE_MINIMUM_MEMORY_BYTES
            )
        )

    if (
        sys.platform == "darwin"
        and object_store_memory > ray_constants.MAC_DEGRADED_PERF_MMAP_SIZE_LIMIT
        and os.environ.get("RAY_ENABLE_MAC_LARGE_OBJECT_STORE") != "1"
    ):
        raise ValueError(
            "The configured object store size ({:.4}GiB) exceeds "
            "the optimal size on Mac ({:.4}GiB). "
            "This will harm performance! There is a known issue where "
            "Ray's performance degrades with object store size greater"
            " than {:.4}GB on a Mac."
            "To reduce the object store capacity, specify"
            "`object_store_memory` when calling ray.init() or ray start."
            "To ignore this warning, "
            "set RAY_ENABLE_MAC_LARGE_OBJECT_STORE=1.".format(
                object_store_memory / 2**30,
                ray_constants.MAC_DEGRADED_PERF_MMAP_SIZE_LIMIT / 2**30,
                ray_constants.MAC_DEGRADED_PERF_MMAP_SIZE_LIMIT / 2**30,
            )
        )

    # Print the object store memory using two decimal places.
    logger.debug(
        "Determine to start the Plasma object store with {} GB memory "
        "using {}.".format(round(object_store_memory / 10**9, 2), plasma_directory)
    )

    return plasma_directory, object_store_memory


def start_monitor(
    gcs_address: str,
    logs_dir: str,
    stdout_file: Optional[str] = None,
    stderr_file: Optional[str] = None,
    autoscaling_config: Optional[str] = None,
    fate_share: Optional[bool] = None,
    max_bytes: int = 0,
    backup_count: int = 0,
    monitor_ip: Optional[str] = None,
):
    """Run a process to monitor the other processes.

    Args:
        gcs_address: The address of GCS server.
        logs_dir: The path to the log directory.
        stdout_file: A file handle opened for writing to redirect stdout to. If
            no redirection should happen, then this should be None.
        stderr_file: A file handle opened for writing to redirect stderr to. If
            no redirection should happen, then this should be None.
        autoscaling_config: path to autoscaling config file.
        max_bytes: Log rotation parameter. Corresponding to
            RotatingFileHandler's maxBytes.
        backup_count: Log rotation parameter. Corresponding to
            RotatingFileHandler's backupCount.
        monitor_ip: IP address of the machine that the monitor will be
            run on. Can be excluded, but required for autoscaler metrics.
    Returns:
        ProcessInfo for the process that was started.
    """
    monitor_path = os.path.join(RAY_PATH, AUTOSCALER_PRIVATE_DIR, "monitor.py")

    command = [
        sys.executable,
        "-u",
        monitor_path,
        f"--logs-dir={logs_dir}",
        f"--logging-rotate-bytes={max_bytes}",
        f"--logging-rotate-backup-count={backup_count}",
    ]
    if gcs_address is not None:
        command.append(f"--gcs-address={gcs_address}")
    if stdout_file is None and stderr_file is None:
        # If not redirecting logging to files, unset log filename.
        # This will cause log records to go to stderr.
        command.append("--logging-filename=")
        # Use stderr log format with the component name as a message prefix.
        logging_format = ray_constants.LOGGER_FORMAT_STDERR.format(
            component=ray_constants.PROCESS_TYPE_MONITOR
        )
        command.append(f"--logging-format={logging_format}")
    if autoscaling_config:
        command.append("--autoscaling-config=" + str(autoscaling_config))
    if monitor_ip:
        command.append("--monitor-ip=" + monitor_ip)
    process_info = start_ray_process(
        command,
        ray_constants.PROCESS_TYPE_MONITOR,
        stdout_file=stdout_file,
        stderr_file=stderr_file,
        fate_share=fate_share,
    )
    return process_info


def start_ray_client_server(
    address: str,
    ray_client_server_ip: str,
    ray_client_server_port: int,
    stdout_file: Optional[int] = None,
    stderr_file: Optional[int] = None,
    redis_password: Optional[int] = None,
    fate_share: Optional[bool] = None,
    metrics_agent_port: Optional[int] = None,
    server_type: str = "proxy",
    serialized_runtime_env_context: Optional[str] = None,
):
    """Run the server process of the Ray client.

    Args:
        address: The address of the cluster.
        ray_client_server_ip: Host IP the Ray client server listens on.
        ray_client_server_port: Port the Ray client server listens on.
        stdout_file: A file handle opened for writing to redirect stdout to. If
            no redirection should happen, then this should be None.
        stderr_file: A file handle opened for writing to redirect stderr to. If
            no redirection should happen, then this should be None.
        redis_password: The password of the redis server.
        server_type: Whether to start the proxy version of Ray Client.
        serialized_runtime_env_context (str|None): If specified, the serialized
            runtime_env_context to start the client server in.

    Returns:
        ProcessInfo for the process that was started.
    """
    root_ray_dir = Path(__file__).resolve().parents[1]
    setup_worker_path = os.path.join(
        root_ray_dir, "_private", "workers", ray_constants.SETUP_WORKER_FILENAME
    )

    ray_client_server_host = (
        "127.0.0.1" if ray_client_server_ip == "127.0.0.1" else "0.0.0.0"
    )
    command = [
        sys.executable,
        setup_worker_path,
        "-m",
        "ray.util.client.server",
        f"--address={address}",
        f"--host={ray_client_server_host}",
        f"--port={ray_client_server_port}",
        f"--mode={server_type}",
        f"--language={Language.Name(Language.PYTHON)}",
    ]
    if redis_password:
        command.append(f"--redis-password={redis_password}")
    if serialized_runtime_env_context:
        command.append(
            f"--serialized-runtime-env-context={serialized_runtime_env_context}"  # noqa: E501
        )
    if metrics_agent_port:
        command.append(f"--metrics-agent-port={metrics_agent_port}")
    process_info = start_ray_process(
        command,
        ray_constants.PROCESS_TYPE_RAY_CLIENT_SERVER,
        stdout_file=stdout_file,
        stderr_file=stderr_file,
        fate_share=fate_share,
    )
    return process_info



"""
Example of interfacing with an environment that produces 2D observations.

This example shows how turning 2D observations with shape (A, B) into a 3D
observations with shape (C, D, 1) can enable usage of RLlib's default models.
RLlib's default Catalog class does not provide default models for 2D observation
spaces, but it does so for 3D observations.
Therefore, one can either write a custom model or transform the 2D observations into 3D
observations. This enables RLlib to use one of the default CNN filters, even though the
original observation space of the environment does not fit them.

This simple example should reach rewards of 50 within 150k timesteps.
"""

from numpy import float32
import argparse
from pettingzoo.butterfly import pistonball_v6
from supersuit import (
    normalize_obs_v0,
    dtype_v0,
    color_reduction_v0,
    reshape_v0,
    resize_v1,
)

from ray.rllib.algorithms.ppo import PPOConfig
from ray.rllib.env import PettingZooEnv
from ray.tune.registry import register_env
from ray import tune
from ray import air


parser = argparse.ArgumentParser()
parser.add_argument(
    "--framework",
    choices=["tf2", "torch"],
    default="torch",
    help="The DL framework specifier.",
)
parser.add_argument(
    "--as-test",
    action="store_true",
    help="Whether this script should be run as a compilation test.",
)
parser.add_argument(
    "--stop-iters", type=int, default=150, help="Number of iterations to train."
)
parser.add_argument(
    "--stop-timesteps", type=int, default=1000000, help="Number of timesteps to train."
)
parser.add_argument(
    "--stop-reward", type=float, default=50, help="Reward at which we stop training."
)

args = parser.parse_args()


# The space we down-sample and transform the greyscale pistonball images to.
# Other spaces supported by RLlib can be chosen here.
TRANSFORMED_OBS_SPACE = (42, 42, 1)


def env_creator(config):
    env = pistonball_v6.env(n_pistons=5)
    env = dtype_v0(env, dtype=float32)
    # This gives us greyscale images for the color red
    env = color_reduction_v0(env, mode="R")
    env = normalize_obs_v0(env)
    # This gives us images that are upsampled to the number of pixels in the
    # default CNN filter
    env = resize_v1(
        env, x_size=TRANSFORMED_OBS_SPACE[0], y_size=TRANSFORMED_OBS_SPACE[1]
    )
    # This gives us 3D images for which we have default filters
    env = reshape_v0(env, shape=TRANSFORMED_OBS_SPACE)
    return env


# Register env
register_env("pistonball", lambda config: PettingZooEnv(env_creator(config)))

config = (
    PPOConfig()
    .environment("pistonball", env_config={"local_ratio": 0.5}, clip_rewards=True)
    .rollouts(
        num_rollout_workers=15 if not args.as_test else 2,
        num_envs_per_worker=1,
        observation_filter="NoFilter",
        rollout_fragment_length="auto",
    )
    .framework("torch")
    .training(
        entropy_coeff=0.01,
        vf_loss_coeff=0.1,
        clip_param=0.1,
        vf_clip_param=10.0,
        num_sgd_iter=10,
        kl_coeff=0.5,
        lr=0.0001,
        grad_clip=100,
        sgd_minibatch_size=500,
        train_batch_size=5000 if not args.as_test else 1000,
        model={"vf_share_layers": True},
    )
    .resources(num_gpus=1 if not args.as_test else 0)
    .reporting(min_time_s_per_iteration=30)
)

tune.Tuner(
    "PPO",
    param_space=config.to_dict(),
    run_config=air.RunConfig(
        stop={
            "training_iteration": args.stop_iters,
            "timesteps_total": args.stop_timesteps,
            "episode_reward_mean": args.stop_reward,
        },
        verbose=2,
    ),
).fit()



from typing import List

from ray.rllib.algorithms.appo.appo_learner import (
    OLD_ACTION_DIST_LOGITS_KEY,
)
from ray.rllib.algorithms.ppo.torch.ppo_torch_rl_module import PPOTorchRLModule
from ray.rllib.core.models.base import ACTOR
from ray.rllib.core.models.tf.encoder import ENCODER_OUT
from ray.rllib.core.rl_module.rl_module_with_target_networks_interface import (
    RLModuleWithTargetNetworksInterface,
)
from ray.rllib.policy.sample_batch import SampleBatch
from ray.rllib.utils.annotations import override
from ray.rllib.utils.nested_dict import NestedDict


class APPOTorchRLModule(PPOTorchRLModule, RLModuleWithTargetNetworksInterface):
    def setup(self):
        super().setup()
        catalog = self.config.get_catalog()
        # Old pi and old encoder are the "target networks" that are used for
        # the stabilization of the updates of the current pi and encoder.
        self.old_pi = catalog.build_pi_head(framework=self.framework)
        self.old_encoder = catalog.build_actor_critic_encoder(framework=self.framework)
        self.old_pi.load_state_dict(self.pi.state_dict())
        self.old_encoder.load_state_dict(self.encoder.state_dict())
        self.old_pi.trainable = False
        self.old_encoder.trainable = False

    @override(RLModuleWithTargetNetworksInterface)
    def get_target_network_pairs(self):
        return [(self.old_pi, self.pi), (self.old_encoder, self.encoder)]

    @override(PPOTorchRLModule)
    def output_specs_train(self) -> List[str]:
        return [
            SampleBatch.ACTION_DIST_INPUTS,
            OLD_ACTION_DIST_LOGITS_KEY,
            SampleBatch.VF_PREDS,
        ]

    @override(PPOTorchRLModule)
    def _forward_train(self, batch: NestedDict):
        outs = super()._forward_train(batch)
        old_pi_inputs_encoded = self.old_encoder(batch)[ENCODER_OUT][ACTOR]
        old_action_dist_logits = self.old_pi(old_pi_inputs_encoded)
        outs[OLD_ACTION_DIST_LOGITS_KEY] = old_action_dist_logits
        return outs
